import fmodpy
lib = fmodpy.fimport("argselect.f90")

print(lib)

import numpy as np

print()
for i in range(100):
    n = 100
    a = np.random.random(size=(n,)).astype("float32")
    k = np.random.randint(1,n)
    i = lib.argselect(a, k) - 1
    v = a[i[k-1]]
    a.sort()
    t = a[k-1]
    assert (v == t), f"\nn: {n}\nk: {k}\ntruth: {t}\nobserved: {v}\na: {a}\ni: {i}"
    if (v != t):
        print(n, k, v == t)



'''This Python code is an automatically generated wrapper
for Fortran code made by 'fmodpy'. The original documentation
for the Fortran source code follows.


'''

import os
import ctypes
import platform
import numpy

# --------------------------------------------------------------------
#               CONFIGURATION
# 
_verbose = True
_fort_compiler = "gfortran"
_shared_object_name = "delsparse." + platform.machine() + ".so"
_this_directory = os.path.dirname(os.path.abspath(__file__))
_path_to_lib = os.path.join(_this_directory, _shared_object_name)
_compile_options = ['-std=legacy', '-fPIC', '-shared', '-O3', '-lblas', '-llapack', '-fopenmp']
_ordered_dependencies = ['slatec.f', 'delsparse.f90', 'delsparse_c_wrapper.f90']
# 
# --------------------------------------------------------------------
#               AUTO-COMPILING
#
# Try to import the existing object. If that fails, recompile and then try.
try:
    clib = ctypes.CDLL(_path_to_lib)
except:
    # Remove the shared object if it exists, because it is faulty.
    if os.path.exists(_shared_object_name):
        os.remove(_shared_object_name)
    # Compile a new shared object.
    _command = " ".join([_fort_compiler] + _compile_options + ["-o", _shared_object_name] + _ordered_dependencies)
    if _verbose:
        print("Running system command with arguments")
        print("  ", _command)
    # Run the compilation command.
    import subprocess
    subprocess.run(_command, shell=True, cwd=_this_directory)
    # Import the shared object file as a C library with ctypes.
    clib = ctypes.CDLL(_path_to_lib)
# --------------------------------------------------------------------


# ----------------------------------------------
# Wrapper for the Fortran subroutine DELAUNAYSPARSES

def delaunaysparses(d, n, pts, m, q, simps, weights, ierr, interp_in=None, interp_out=None, eps=None, extrap=None, rnorm=None, ibudget=None, chain=None):
    '''! This is a serial implementation of an algorithm for efficiently performing
! interpolation in R^D via the Delaunay triangulation. The algorithm is fully
! described and analyzed in
!
! T. H. Chang, L. T. Watson,  T. C.H. Lux, B. Li, L. Xu, A. R. Butt, K. W.
! Cameron, and Y. Hong. 2018. A polynomial time algorithm for multivariate
! interpolation in arbitrary dimension via the Delaunay triangulation. In
! Proceedings of the ACMSE 2018 Conference (ACMSE '18). ACM, New York, NY,
! USA. Article 12, 8 pages.
!
!
! On input:
!
! D is the dimension of the space for PTS and Q.
!
! N is the number of data points in PTS.
!
! PTS(1:D,1:N) is a real valued matrix with N columns, each containing the
!    coordinates of a single data point in R^D.
!
! M is the number of interpolation points in Q.
!
! Q(1:D,1:M) is a real valued matrix with M columns, each containing the
!    coordinates of a single interpolation point in R^D.
!
!
! On output:
!
! PTS and Q have been rescaled and shifted. All the data points in PTS
!    are now contained in the unit hyperball in R^D, and the points in Q
!    have been shifted and scaled accordingly in relation to PTS.
!
! SIMPS(1:D+1,1:M) contains the D+1 integer indices (corresponding to columns
!    in PTS) for the D+1 vertices of the Delaunay simplex containing each
!    interpolation point in Q.
!
! WEIGHTS(1:D+1,1:M) contains the D+1 real valued weights for expressing each
!    point in Q as a convex combination of the D+1 corresponding vertices
!    in SIMPS.
!
! IERR(1:M) contains integer valued error flags associated with the
!    computation of each of the M interpolation points in Q. The error
!    codes are:
!
! 00 : Succesful interpolation.
! 01 : Succesful extrapolation (up to the allowed extrapolation distance).
! 02 : This point was outside the allowed extrapolation distance; the
!      corresponding entries in SIMPS and WEIGHTS contain zero values.
!
! 10 : The dimension D must be positive.
! 11 : Too few data points to construct a triangulation (i.e., N < D+1).
! 12 : No interpolation points given (i.e., M < 1).
! 13 : The first dimension of PTS does not agree with the dimension D.
! 14 : The second dimension of PTS does not agree with the number of points N.
! 15 : The first dimension of Q does not agree with the dimension D.
! 16 : The second dimension of Q does not agree with the number of
!      interpolation points M.
! 17 : The first dimension of the output array SIMPS does not match the number
!      of vertices needed for a D-simplex (D+1).
! 18 : The second dimension of the output array SIMPS does not match the
!      number of interpolation points M.
! 19 : The first dimension of the output array WEIGHTS does not match the
!      number of vertices for a a D-simplex (D+1).
! 20 : The second dimension of the output array WEIGHTS does not match the
!      number of interpolation points M.
! 21 : The size of the error array IERR does not match the number of
!      interpolation points M.
! 22 : INTERP_IN cannot be present without INTERP_OUT or vice versa.
! 23 : The first dimension of INTERP_IN does not match the first
!      dimension of INTERP_OUT.
! 24 : The second dimension of INTERP_IN does not match the number of
!      data points PTS.
! 25 : The second dimension of INTERP_OUT does not match the number of
!      interpolation points M.
! 26 : The budget supplied in IBUDGET does not contain a positive
!      integer.
! 27 : The extrapolation distance supplied in EXTRAP cannot be negative.
! 28 : The size of the RNORM output array does not match the number of
!      interpolation points M.
!
! 30 : Two or more points in the data set PTS are too close together with
!      respect to the working precision (EPS), which would result in a
!      numerically degenerate simplex.
! 31 : All the data points in PTS lie in some lower dimensional linear
!      manifold (up to the working precision), and no valid triangulation
!      exists.
! 40 : An error caused DELAUNAYSPARSES to terminate before this value could
!      be computed. Note: The corresponding entries in SIMPS and WEIGHTS may
!      contain garbage values.
!
! 50 : A memory allocation error occurred while allocating the work array
!      WORK.
!
! 60 : The budget was exceeded before the algorithm converged on this
!      value. If the dimension is high, try increasing IBUDGET. This
!      error can also be caused by a working precision EPS that is too
!      small for the conditioning of the problem.
!
! 61 : A value that was judged appropriate later caused LAPACK to encounter a
!      singularity. Try increasing the value of EPS.
!
! 70 : Allocation error for the extrapolation work arrays.
! 71 : The SLATEC subroutine DWNNLS failed to converge during the projection
!      of an extrapolation point onto the convex hull.
! 72 : The SLATEC subroutine DWNNLS has reported a usage error.
!
!      The errors 72, 80--83 should never occur, and likely indicate a
!      compiler bug or hardware failure.
! 80 : The LAPACK subroutine DGEQP3 has reported an illegal value.
! 81 : The LAPACK subroutine DGETRF has reported an illegal value.
! 82 : The LAPACK subroutine DGETRS has reported an illegal value.
! 83 : The LAPACK subroutine DORMQR has reported an illegal value.
!
! Optional arguments:
!
! INTERP_IN(1:IR,1:N) contains real valued response vectors for each of
!    the data points in PTS on input. The first dimension of INTERP_IN is
!    inferred to be the dimension of these response vectors, and the
!    second dimension must match N. If present, the response values will
!    be computed for each interpolation point in Q, and stored in INTERP_OUT,
!    which therefore must also be present. If both INTERP_IN and INTERP_OUT
!    are omitted, only the containing simplices and convex combination
!    weights are returned.
!
! INTERP_OUT(1:IR,1:M) contains real valued response vectors for each
!    interpolation point in Q on output. The first dimension of INTERP_OUT
!    must match the first dimension of INTERP_IN, and the second dimension
!    must match M. If present, the response values at each interpolation
!    point are computed as a convex combination of the response values
!    (supplied in INTERP_IN) at the vertices of a Delaunay simplex containing
!    that interpolation point.  Therefore, if INTERP_OUT is present, then
!    INTERP_IN must also be present.  If both are omitted, only the
!    simplices and convex combination weights are returned.
!
! EPS contains the real working precision for the problem on input. By default,
!    EPS is assigned \sqrt{\mu} where \mu denotes the unit roundoff for the
!    machine. In general, any values that differ by less than EPS are judged
!    as equal, and any weights that are greater than -EPS are judged as
!    nonnegative.  EPS cannot take a value less than the default value of
!    \sqrt{\mu}. If any value less than \sqrt{\mu} is supplied, the default
!    value will be used instead automatically.
!
! EXTRAP contains the real maximum extrapolation distance (relative to the
!    diameter of PTS) on input. Interpolation at a point outside the convex
!    hull of PTS is done by projecting that point onto the convex hull, and
!    then doing normal Delaunay interpolation at that projection.
!    Interpolation at any point in Q that is more than EXTRAP * DIAMETER(PTS)
!    units outside the convex hull of PTS will not be done and an error code
!    of 2 will be returned. Note that computing the projection can be
!    expensive. Setting EXTRAP=0 will cause all extrapolation points to be
!    ignored without ever computing a projection. By default, EXTRAP=0.1
!    (extrapolate by up to 10% of the diameter of PTS).
!
! RNORM(1:M) contains the real unscaled projection (2-norm) distances from
!    any projection computations on output. If not present, these distances
!    are still computed for each extrapolation point, but are never returned.
!
! IBUDGET on input contains the integer budget for performing flips while
!    iterating toward the simplex containing each interpolation point in
!    Q. This prevents DELAUNAYSPARSES from falling into an infinite loop when
!    an inappropriate value of EPS is given with respect to the problem
!    conditioning.  By default, IBUDGET=50000. However, for extremely
!    high-dimensional problems and pathological inputs, the default value
!    may be insufficient.
!
! CHAIN is a logical input argument that determines whether a new first
!    simplex should be constructed for each interpolation point
!    (CHAIN=.FALSE.), or whether the simplex walks should be "daisy-chained."
!    By default, CHAIN=.FALSE. Setting CHAIN=.TRUE. is generally not
!    recommended, unless the size of the triangulation is relatively small
!    or the interpolation points are known to be tightly clustered.
!
!
! Subroutines and functions directly referenced from BLAS are
!      DDOT, DGEMV, DNRM2, DTRSM,
! and from LAPACK are
!      DGEQP3, DGETRF, DGETRS, DORMQR.
! The SLATEC subroutine DWNNLS is directly referenced. DWNNLS and all its
! SLATEC dependencies have been slightly edited to comply with the Fortran
! 2008 standard, with all print statements and references to stderr being
! commented out. For a reference to DWNNLS, see ACM TOMS Algorithm 587
! (Hanson and Haskell).  The module REAL_PRECISION from HOMPACK90 (ACM TOMS
! Algorithm 777) is used for the real data type. The REAL_PRECISION module,
! DELAUNAYSPARSES, and DWNNLS and its dependencies comply with the Fortran
! 2008 standard.
!
! Primary Author: Tyler H. Chang
! Last Update: February, 2019'''
    
    # Setting up "d"
    if (type(d) is not ctypes.c_int): d = ctypes.c_int(d)
    
    # Setting up "n"
    if (type(n) is not ctypes.c_int): n = ctypes.c_int(n)
    
    # Setting up "pts"
    if ((not issubclass(type(pts), numpy.ndarray)) or
        (not numpy.asarray(pts).flags.f_contiguous) or
        (not (pts.dtype == numpy.dtype(ctypes.c_double)))):
        import warnings
        warnings.warn("The provided argument 'pts' was not an f_contiguous NumPy array of type 'ctypes.c_double' (or equivalent). Automatically converting (probably creating a full copy).")
        pts = numpy.asarray(pts, dtype=ctypes.c_double, order='F')
    pts_dim_1 = ctypes.c_int(pts.shape[0])
    pts_dim_2 = ctypes.c_int(pts.shape[1])
    
    # Setting up "m"
    if (type(m) is not ctypes.c_int): m = ctypes.c_int(m)
    
    # Setting up "q"
    if ((not issubclass(type(q), numpy.ndarray)) or
        (not numpy.asarray(q).flags.f_contiguous) or
        (not (q.dtype == numpy.dtype(ctypes.c_double)))):
        import warnings
        warnings.warn("The provided argument 'q' was not an f_contiguous NumPy array of type 'ctypes.c_double' (or equivalent). Automatically converting (probably creating a full copy).")
        q = numpy.asarray(q, dtype=ctypes.c_double, order='F')
    q_dim_1 = ctypes.c_int(q.shape[0])
    q_dim_2 = ctypes.c_int(q.shape[1])
    
    # Setting up "simps"
    if ((not issubclass(type(simps), numpy.ndarray)) or
        (not numpy.asarray(simps).flags.f_contiguous) or
        (not (simps.dtype == numpy.dtype(ctypes.c_int)))):
        import warnings
        warnings.warn("The provided argument 'simps' was not an f_contiguous NumPy array of type 'ctypes.c_int' (or equivalent). Automatically converting (probably creating a full copy).")
        simps = numpy.asarray(simps, dtype=ctypes.c_int, order='F')
    simps_dim_1 = ctypes.c_int(simps.shape[0])
    simps_dim_2 = ctypes.c_int(simps.shape[1])
    
    # Setting up "weights"
    if ((not issubclass(type(weights), numpy.ndarray)) or
        (not numpy.asarray(weights).flags.f_contiguous) or
        (not (weights.dtype == numpy.dtype(ctypes.c_double)))):
        import warnings
        warnings.warn("The provided argument 'weights' was not an f_contiguous NumPy array of type 'ctypes.c_double' (or equivalent). Automatically converting (probably creating a full copy).")
        weights = numpy.asarray(weights, dtype=ctypes.c_double, order='F')
    weights_dim_1 = ctypes.c_int(weights.shape[0])
    weights_dim_2 = ctypes.c_int(weights.shape[1])
    
    # Setting up "ierr"
    if ((not issubclass(type(ierr), numpy.ndarray)) or
        (not numpy.asarray(ierr).flags.f_contiguous) or
        (not (ierr.dtype == numpy.dtype(ctypes.c_int)))):
        import warnings
        warnings.warn("The provided argument 'ierr' was not an f_contiguous NumPy array of type 'ctypes.c_int' (or equivalent). Automatically converting (probably creating a full copy).")
        ierr = numpy.asarray(ierr, dtype=ctypes.c_int, order='F')
    ierr_dim_1 = ctypes.c_int(ierr.shape[0])
    
    # Setting up "interp_in"
    interp_in_present = ctypes.c_bool(True)
    if (interp_in is None):
        interp_in_present = ctypes.c_bool(False)
        interp_in = numpy.zeros(shape=(1,1), dtype=ctypes.c_double, order='F')
    elif (type(interp_in) == bool) and (interp_in):
        interp_in = numpy.zeros(shape=(1,1), dtype=ctypes.c_double, order='F')
    elif ((not issubclass(type(interp_in), numpy.ndarray)) or
          (not numpy.asarray(interp_in).flags.f_contiguous) or
          (not (interp_in.dtype == numpy.dtype(ctypes.c_double)))):
        import warnings
        warnings.warn("The provided argument 'interp_in' was not an f_contiguous NumPy array of type 'ctypes.c_double' (or equivalent). Automatically converting (probably creating a full copy).")
        interp_in = numpy.asarray(interp_in, dtype=ctypes.c_double, order='F')
    if (interp_in_present):
        interp_in_dim_1 = ctypes.c_int(interp_in.shape[0])
        interp_in_dim_2 = ctypes.c_int(interp_in.shape[1])
    else:
        interp_in_dim_1 = ctypes.c_int()
        interp_in_dim_2 = ctypes.c_int()
    
    # Setting up "interp_out"
    interp_out_present = ctypes.c_bool(True)
    if (interp_out is None):
        interp_out_present = ctypes.c_bool(False)
        interp_out = numpy.zeros(shape=(1,1), dtype=ctypes.c_double, order='F')
    elif (type(interp_out) == bool) and (interp_out):
        interp_out = numpy.zeros(shape=(1,1), dtype=ctypes.c_double, order='F')
    elif ((not issubclass(type(interp_out), numpy.ndarray)) or
          (not numpy.asarray(interp_out).flags.f_contiguous) or
          (not (interp_out.dtype == numpy.dtype(ctypes.c_double)))):
        import warnings
        warnings.warn("The provided argument 'interp_out' was not an f_contiguous NumPy array of type 'ctypes.c_double' (or equivalent). Automatically converting (probably creating a full copy).")
        interp_out = numpy.asarray(interp_out, dtype=ctypes.c_double, order='F')
    if (interp_out_present):
        interp_out_dim_1 = ctypes.c_int(interp_out.shape[0])
        interp_out_dim_2 = ctypes.c_int(interp_out.shape[1])
    else:
        interp_out_dim_1 = ctypes.c_int()
        interp_out_dim_2 = ctypes.c_int()
    
    # Setting up "eps"
    eps_present = ctypes.c_bool(True)
    if (eps is None):
        eps_present = ctypes.c_bool(False)
        eps = ctypes.c_double()
    if (type(eps) is not ctypes.c_double): eps = ctypes.c_double(eps)
    
    # Setting up "extrap"
    extrap_present = ctypes.c_bool(True)
    if (extrap is None):
        extrap_present = ctypes.c_bool(False)
        extrap = ctypes.c_double()
    if (type(extrap) is not ctypes.c_double): extrap = ctypes.c_double(extrap)
    
    # Setting up "rnorm"
    rnorm_present = ctypes.c_bool(True)
    if (rnorm is None):
        rnorm_present = ctypes.c_bool(False)
        rnorm = numpy.zeros(shape=(1), dtype=ctypes.c_double, order='F')
    elif (type(rnorm) == bool) and (rnorm):
        rnorm = numpy.zeros(shape=(1), dtype=ctypes.c_double, order='F')
    elif ((not issubclass(type(rnorm), numpy.ndarray)) or
          (not numpy.asarray(rnorm).flags.f_contiguous) or
          (not (rnorm.dtype == numpy.dtype(ctypes.c_double)))):
        import warnings
        warnings.warn("The provided argument 'rnorm' was not an f_contiguous NumPy array of type 'ctypes.c_double' (or equivalent). Automatically converting (probably creating a full copy).")
        rnorm = numpy.asarray(rnorm, dtype=ctypes.c_double, order='F')
    if (rnorm_present):
        rnorm_dim_1 = ctypes.c_int(rnorm.shape[0])
    else:
        rnorm_dim_1 = ctypes.c_int()
    
    # Setting up "ibudget"
    ibudget_present = ctypes.c_bool(True)
    if (ibudget is None):
        ibudget_present = ctypes.c_bool(False)
        ibudget = ctypes.c_int()
    if (type(ibudget) is not ctypes.c_int): ibudget = ctypes.c_int(ibudget)
    
    # Setting up "chain"
    chain_present = ctypes.c_bool(True)
    if (chain is None):
        chain_present = ctypes.c_bool(False)
        chain = ctypes.c_bool()
    if (type(chain) is not ctypes.c_bool): chain = ctypes.c_bool(chain)

    # Call C-accessible Fortran wrapper.
    clib.c_delaunaysparses(ctypes.byref(d), ctypes.byref(n), ctypes.byref(pts_dim_1), ctypes.byref(pts_dim_2), ctypes.c_void_p(pts.ctypes.data), ctypes.byref(m), ctypes.byref(q_dim_1), ctypes.byref(q_dim_2), ctypes.c_void_p(q.ctypes.data), ctypes.byref(simps_dim_1), ctypes.byref(simps_dim_2), ctypes.c_void_p(simps.ctypes.data), ctypes.byref(weights_dim_1), ctypes.byref(weights_dim_2), ctypes.c_void_p(weights.ctypes.data), ctypes.byref(ierr_dim_1), ctypes.c_void_p(ierr.ctypes.data), ctypes.byref(interp_in_present), ctypes.byref(interp_in_dim_1), ctypes.byref(interp_in_dim_2), ctypes.c_void_p(interp_in.ctypes.data), ctypes.byref(interp_out_present), ctypes.byref(interp_out_dim_1), ctypes.byref(interp_out_dim_2), ctypes.c_void_p(interp_out.ctypes.data), ctypes.byref(eps_present), ctypes.byref(eps), ctypes.byref(extrap_present), ctypes.byref(extrap), ctypes.byref(rnorm_present), ctypes.byref(rnorm_dim_1), ctypes.c_void_p(rnorm.ctypes.data), ctypes.byref(ibudget_present), ctypes.byref(ibudget), ctypes.byref(chain_present), ctypes.byref(chain))

    # Return final results, 'INTENT(OUT)' arguments only.
    return pts, q, simps, weights, ierr, (interp_out if interp_out_present else None), (rnorm if rnorm_present else None)


# ----------------------------------------------
# Wrapper for the Fortran subroutine DELAUNAYSPARSEP

def delaunaysparsep(d, n, pts, m, q, simps, weights, ierr, interp_in=None, interp_out=None, eps=None, extrap=None, rnorm=None, ibudget=None, chain=None, pmode=None):
    '''! This is a parallel implementation of an algorithm for efficiently performing
! interpolation in R^D via the Delaunay triangulation. The algorithm is fully
! described and analyzed in
!
! T. H. Chang, L. T. Watson,  T. C.H. Lux, B. Li, L. Xu, A. R. Butt, K. W.
! Cameron, and Y. Hong. 2018. A polynomial time algorithm for multivariate
! interpolation in arbitrary dimension via the Delaunay triangulation. In
! Proceedings of the ACMSE 2018 Conference (ACMSE '18). ACM, New York, NY,
! USA. Article 12, 8 pages.
!
!
! On input:
!
! D is the dimension of the space for PTS and Q.
!
! N is the number of data points in PTS.
!
! PTS(1:D,1:N) is a real valued matrix with N columns, each containing the
!    coordinates of a single data point in R^D.
!
! M is the number of interpolation points in Q.
!
! Q(1:D,1:M) is a real valued matrix with M columns, each containing the
!    coordinates of a single interpolation point in R^D.
!
!
! On output:
!
! PTS and Q have been rescaled and shifted. All the data points in PTS
!    are now contained in the unit hyperball in R^D, and the points in Q
!    have been shifted and scaled accordingly in relation to PTS.
!
! SIMPS(1:D+1,1:M) contains the D+1 integer indices (corresponding to columns
!    in PTS) for the D+1 vertices of the Delaunay simplex containing each
!    interpolation point in Q.
!
! WEIGHTS(1:D+1,1:M) contains the D+1 real valued weights for expressing each
!    point in Q as a convex combination of the D+1 corresponding vertices
!    in SIMPS.
!
! IERR(1:M) contains integer valued error flags associated with the
!    computation of each of the M interpolation points in Q. The error
!    codes are:
!
! 00 : Succesful interpolation.
! 01 : Succesful extrapolation (up to the allowed extrapolation distance).
! 02 : This point was outside the allowed extrapolation distance; the
!      corresponding entries in SIMPS and WEIGHTS contain zero values.
!
! 10 : The dimension D must be positive.
! 11 : Too few data points to construct a triangulation (i.e., N < D+1).
! 12 : No interpolation points given (i.e., M < 1).
! 13 : The first dimension of PTS does not agree with the dimension D.
! 14 : The second dimension of PTS does not agree with the number of points N.
! 15 : The first dimension of Q does not agree with the dimension D.
! 16 : The second dimension of Q does not agree with the number of
!      interpolation points M.
! 17 : The first dimension of the output array SIMPS does not match the number
!      of vertices needed for a D-simplex (D+1).
! 18 : The second dimension of the output array SIMPS does not match the
!      number of interpolation points M.
! 19 : The first dimension of the output array WEIGHTS does not match the
!      number of vertices for a a D-simplex (D+1).
! 20 : The second dimension of the output array WEIGHTS does not match the
!      number of interpolation points M.
! 21 : The size of the error array IERR does not match the number of
!      interpolation points M.
! 22 : INTERP_IN cannot be present without INTERP_OUT or vice versa.
! 23 : The first dimension of INTERP_IN does not match the first
!      dimension of INTERP_OUT.
! 24 : The second dimension of INTERP_IN does not match the number of
!      data points PTS.
! 25 : The second dimension of INTERP_OUT does not match the number of
!      interpolation points M.
! 26 : The budget supplied in IBUDGET does not contain a positive
!      integer.
! 27 : The extrapolation distance supplied in EXTRAP cannot be negative.
! 28 : The size of the RNORM output array does not match the number of
!      interpolation points M.
!
! 30 : Two or more points in the data set PTS are too close together with
!      respect to the working precision (EPS), which would result in a
!      numerically degenerate simplex.
! 31 : All the data points in PTS lie in some lower dimensional linear
!      manifold (up to the working precision), and no valid triangulation
!      exists.
! 40 : An error caused DELAUNAYSPARSEP to terminate before this value could
!      be computed. Note: The corresponding entries in SIMPS and WEIGHTS may
!      contain garbage values.
!
! 50 : A memory allocation error occurred while allocating the work array
!      WORK.
!
! 60 : The budget was exceeded before the algorithm converged on this
!      value. If the dimension is high, try increasing IBUDGET. This
!      error can also be caused by a working precision EPS that is too
!      small for the conditioning of the problem.
!
! 61 : A value that was judged appropriate later caused LAPACK to encounter a
!      singularity. Try increasing the value of EPS.
!
! 70 : Allocation error for the extrapolation work arrays.
! 71 : The SLATEC subroutine DWNNLS failed to converge during the projection
!      of an extrapolation point onto the convex hull.
! 72 : The SLATEC subroutine DWNNLS has reported a usage error.
!
!      The errors 72, 80--83 should never occur, and likely indicate a
!      compiler bug or hardware failure.
! 80 : The LAPACK subroutine DGEQP3 has reported an illegal value.
! 81 : The LAPACK subroutine DGETRF has reported an illegal value.
! 82 : The LAPACK subroutine DGETRS has reported an illegal value.
! 83 : The LAPACK subroutine DORMQR has reported an illegal value.
!
! 90 : The value of PMODE is not valid.
!
! Optional arguments:
!
! INTERP_IN(1:IR,1:N) contains real valued response vectors for each of
!    the data points in PTS on input. The first dimension of INTERP_IN is
!    inferred to be the dimension of these response vectors, and the
!    second dimension must match N. If present, the response values will
!    be computed for each interpolation point in Q, and stored in INTERP_OUT,
!    which therefore must also be present. If both INTERP_IN and INTERP_OUT
!    are omitted, only the containing simplices and convex combination
!    weights are returned.
!
! INTERP_OUT(1:IR,1:M) contains real valued response vectors for each
!    interpolation point in Q on output. The first dimension of INTERP_OU
!    must match the first dimension of INTERP_IN, and the second dimension
!    must match M. If present, the response values at each interpolation
!    point are computed as a convex combination of the response values
!    (supplied in INTERP_IN) at the vertices of a Delaunay simplex containing
!    that interpolation point.  Therefore, if INTERP_OUT is present, then
!    INTERP_IN must also be present.  If both are omitted, only the
!    simplices and convex combination weights are returned.
!
! EPS contains the real working precision for the problem on input. By
!    default, EPS is assigned \sqrt{\mu} where \mu denotes the unit roundoff
!    for the machine. In general, any values that differ by less than EPS
!    are judged as equal, and any weights that are greater than -EPS are
!    judged as nonnegative.  EPS cannot take a value less than the default
!    value of \sqrt{\mu}. If any value less than \sqrt{\mu} is supplied,
!    the default value will be used instead automatically.
!
! EXTRAP contains the real maximum extrapolation distance (relative to the
!    diameter of PTS) on input. Interpolation at a point outside the convex
!    hull of PTS is done by projecting that point onto the convex hull, and
!    then doing normal Delaunay interpolation at that projection.
!    Interpolation at any point in Q that is more than EXTRAP * DIAMETER(PTS)
!    units outside the convex hull of PTS will not be done and an error code
!    of 2 will be returned. Note that computing the projection can be
!    expensive. Setting EXTRAP=0 will cause all extrapolation points to be
!    ignored without ever computing a projection. By default, EXTRAP=0.1
!    (extrapolate by up to 10% of the diameter of PTS).
!
! RNORM(1:M) contains the real unscaled projection (2-norm) distances from
!    any projection computations on output. If not present, these distances
!    are still computed for each extrapolation point, but are never returned.
!
! IBUDGET on input contains the integer budget for performing flips while
!    iterating toward the simplex containing each interpolation point in Q.
!    This prevents DELAUNAYSPARSEP from falling into an infinite loop when
!    an inappropriate value of EPS is given with respect to the problem
!    conditioning.  By default, IBUDGET=50000. However, for extremely
!    high-dimensional problems and pathological inputs, the default value
!    may be insufficient.
!
! CHAIN is a logical input argument that determines whether a new first
!    simplex should be constructed for each interpolation point
!    (CHAIN=.FALSE.), or whether the simplex walks should be "daisy-chained."
!    By default, CHAIN=.FALSE. Setting CHAIN=.TRUE. is generally not
!    recommended, unless the size of the triangulation is relatively small
!    or the interpolation points are known to be tightly clustered.
!
! PMODE is an integer specifying the level of parallelism to be exploited.
!    If PMODE = 1, then parallelism is exploited at the level of the loop
!    over all interpolation points (Level 1 parallelism).
!    If PMODE = 2, then parallelism is exploited at the level of the loops
!    over data points when constructing/flipping simplices (Level 2
!    parallelism).
!    If PMODE = 3, then parallelism is exploited at both levels. Note: this
!    implies that the total number of threads active at any time could be up
!    to OMP_NUM_THREADS^2.
!    By default, PMODE is set to 1 if there is more than 1 interpolation
!    point and 2 otherwise.
!
! Subroutines and functions directly referenced from BLAS are
!      DDOT, DGEMV, DNRM2, DTRSM,
! and from LAPACK are
!      DGEQP3, DGETRF, DGETRS, DORMQR.
! The SLATEC subroutine DWNNLS is directly referenced. DWNNLS and all its
! SLATEC dependencies have been slightly edited to comply with the Fortran
! 2008 standard, with all print statements and references to stderr being
! commented out. For a reference to DWNNLS, see ACM TOMS Algorithm 587
! (Hanson and Haskell).  The module REAL_PRECISION from HOMPACK90 (ACM TOMS
! Algorithm 777) is used for the real data type. The REAL_PRECISION module,
! DELAUNAYSPARSEP, and DWNNLS and its dependencies comply with the Fortran
! 2008 standard.
!
! Primary Author: Tyler H. Chang
! Last Update: February, 2019'''
    
    # Setting up "d"
    if (type(d) is not ctypes.c_int): d = ctypes.c_int(d)
    
    # Setting up "n"
    if (type(n) is not ctypes.c_int): n = ctypes.c_int(n)
    
    # Setting up "pts"
    if ((not issubclass(type(pts), numpy.ndarray)) or
        (not numpy.asarray(pts).flags.f_contiguous) or
        (not (pts.dtype == numpy.dtype(ctypes.c_double)))):
        import warnings
        warnings.warn("The provided argument 'pts' was not an f_contiguous NumPy array of type 'ctypes.c_double' (or equivalent). Automatically converting (probably creating a full copy).")
        pts = numpy.asarray(pts, dtype=ctypes.c_double, order='F')
    pts_dim_1 = ctypes.c_int(pts.shape[0])
    pts_dim_2 = ctypes.c_int(pts.shape[1])
    
    # Setting up "m"
    if (type(m) is not ctypes.c_int): m = ctypes.c_int(m)
    
    # Setting up "q"
    if ((not issubclass(type(q), numpy.ndarray)) or
        (not numpy.asarray(q).flags.f_contiguous) or
        (not (q.dtype == numpy.dtype(ctypes.c_double)))):
        import warnings
        warnings.warn("The provided argument 'q' was not an f_contiguous NumPy array of type 'ctypes.c_double' (or equivalent). Automatically converting (probably creating a full copy).")
        q = numpy.asarray(q, dtype=ctypes.c_double, order='F')
    q_dim_1 = ctypes.c_int(q.shape[0])
    q_dim_2 = ctypes.c_int(q.shape[1])
    
    # Setting up "simps"
    if ((not issubclass(type(simps), numpy.ndarray)) or
        (not numpy.asarray(simps).flags.f_contiguous) or
        (not (simps.dtype == numpy.dtype(ctypes.c_int)))):
        import warnings
        warnings.warn("The provided argument 'simps' was not an f_contiguous NumPy array of type 'ctypes.c_int' (or equivalent). Automatically converting (probably creating a full copy).")
        simps = numpy.asarray(simps, dtype=ctypes.c_int, order='F')
    simps_dim_1 = ctypes.c_int(simps.shape[0])
    simps_dim_2 = ctypes.c_int(simps.shape[1])
    
    # Setting up "weights"
    if ((not issubclass(type(weights), numpy.ndarray)) or
        (not numpy.asarray(weights).flags.f_contiguous) or
        (not (weights.dtype == numpy.dtype(ctypes.c_double)))):
        import warnings
        warnings.warn("The provided argument 'weights' was not an f_contiguous NumPy array of type 'ctypes.c_double' (or equivalent). Automatically converting (probably creating a full copy).")
        weights = numpy.asarray(weights, dtype=ctypes.c_double, order='F')
    weights_dim_1 = ctypes.c_int(weights.shape[0])
    weights_dim_2 = ctypes.c_int(weights.shape[1])
    
    # Setting up "ierr"
    if ((not issubclass(type(ierr), numpy.ndarray)) or
        (not numpy.asarray(ierr).flags.f_contiguous) or
        (not (ierr.dtype == numpy.dtype(ctypes.c_int)))):
        import warnings
        warnings.warn("The provided argument 'ierr' was not an f_contiguous NumPy array of type 'ctypes.c_int' (or equivalent). Automatically converting (probably creating a full copy).")
        ierr = numpy.asarray(ierr, dtype=ctypes.c_int, order='F')
    ierr_dim_1 = ctypes.c_int(ierr.shape[0])
    
    # Setting up "interp_in"
    interp_in_present = ctypes.c_bool(True)
    if (interp_in is None):
        interp_in_present = ctypes.c_bool(False)
        interp_in = numpy.zeros(shape=(1,1), dtype=ctypes.c_double, order='F')
    elif (type(interp_in) == bool) and (interp_in):
        interp_in = numpy.zeros(shape=(1,1), dtype=ctypes.c_double, order='F')
    elif ((not issubclass(type(interp_in), numpy.ndarray)) or
          (not numpy.asarray(interp_in).flags.f_contiguous) or
          (not (interp_in.dtype == numpy.dtype(ctypes.c_double)))):
        import warnings
        warnings.warn("The provided argument 'interp_in' was not an f_contiguous NumPy array of type 'ctypes.c_double' (or equivalent). Automatically converting (probably creating a full copy).")
        interp_in = numpy.asarray(interp_in, dtype=ctypes.c_double, order='F')
    if (interp_in_present):
        interp_in_dim_1 = ctypes.c_int(interp_in.shape[0])
        interp_in_dim_2 = ctypes.c_int(interp_in.shape[1])
    else:
        interp_in_dim_1 = ctypes.c_int()
        interp_in_dim_2 = ctypes.c_int()
    
    # Setting up "interp_out"
    interp_out_present = ctypes.c_bool(True)
    if (interp_out is None):
        interp_out_present = ctypes.c_bool(False)
        interp_out = numpy.zeros(shape=(1,1), dtype=ctypes.c_double, order='F')
    elif (type(interp_out) == bool) and (interp_out):
        interp_out = numpy.zeros(shape=(1,1), dtype=ctypes.c_double, order='F')
    elif ((not issubclass(type(interp_out), numpy.ndarray)) or
          (not numpy.asarray(interp_out).flags.f_contiguous) or
          (not (interp_out.dtype == numpy.dtype(ctypes.c_double)))):
        import warnings
        warnings.warn("The provided argument 'interp_out' was not an f_contiguous NumPy array of type 'ctypes.c_double' (or equivalent). Automatically converting (probably creating a full copy).")
        interp_out = numpy.asarray(interp_out, dtype=ctypes.c_double, order='F')
    if (interp_out_present):
        interp_out_dim_1 = ctypes.c_int(interp_out.shape[0])
        interp_out_dim_2 = ctypes.c_int(interp_out.shape[1])
    else:
        interp_out_dim_1 = ctypes.c_int()
        interp_out_dim_2 = ctypes.c_int()
    
    # Setting up "eps"
    eps_present = ctypes.c_bool(True)
    if (eps is None):
        eps_present = ctypes.c_bool(False)
        eps = ctypes.c_double()
    if (type(eps) is not ctypes.c_double): eps = ctypes.c_double(eps)
    
    # Setting up "extrap"
    extrap_present = ctypes.c_bool(True)
    if (extrap is None):
        extrap_present = ctypes.c_bool(False)
        extrap = ctypes.c_double()
    if (type(extrap) is not ctypes.c_double): extrap = ctypes.c_double(extrap)
    
    # Setting up "rnorm"
    rnorm_present = ctypes.c_bool(True)
    if (rnorm is None):
        rnorm_present = ctypes.c_bool(False)
        rnorm = numpy.zeros(shape=(1), dtype=ctypes.c_double, order='F')
    elif (type(rnorm) == bool) and (rnorm):
        rnorm = numpy.zeros(shape=(1), dtype=ctypes.c_double, order='F')
    elif ((not issubclass(type(rnorm), numpy.ndarray)) or
          (not numpy.asarray(rnorm).flags.f_contiguous) or
          (not (rnorm.dtype == numpy.dtype(ctypes.c_double)))):
        import warnings
        warnings.warn("The provided argument 'rnorm' was not an f_contiguous NumPy array of type 'ctypes.c_double' (or equivalent). Automatically converting (probably creating a full copy).")
        rnorm = numpy.asarray(rnorm, dtype=ctypes.c_double, order='F')
    if (rnorm_present):
        rnorm_dim_1 = ctypes.c_int(rnorm.shape[0])
    else:
        rnorm_dim_1 = ctypes.c_int()
    
    # Setting up "ibudget"
    ibudget_present = ctypes.c_bool(True)
    if (ibudget is None):
        ibudget_present = ctypes.c_bool(False)
        ibudget = ctypes.c_int()
    if (type(ibudget) is not ctypes.c_int): ibudget = ctypes.c_int(ibudget)
    
    # Setting up "chain"
    chain_present = ctypes.c_bool(True)
    if (chain is None):
        chain_present = ctypes.c_bool(False)
        chain = ctypes.c_bool()
    if (type(chain) is not ctypes.c_bool): chain = ctypes.c_bool(chain)
    
    # Setting up "pmode"
    pmode_present = ctypes.c_bool(True)
    if (pmode is None):
        pmode_present = ctypes.c_bool(False)
        pmode = ctypes.c_int()
    if (type(pmode) is not ctypes.c_int): pmode = ctypes.c_int(pmode)

    # Call C-accessible Fortran wrapper.
    clib.c_delaunaysparsep(ctypes.byref(d), ctypes.byref(n), ctypes.byref(pts_dim_1), ctypes.byref(pts_dim_2), ctypes.c_void_p(pts.ctypes.data), ctypes.byref(m), ctypes.byref(q_dim_1), ctypes.byref(q_dim_2), ctypes.c_void_p(q.ctypes.data), ctypes.byref(simps_dim_1), ctypes.byref(simps_dim_2), ctypes.c_void_p(simps.ctypes.data), ctypes.byref(weights_dim_1), ctypes.byref(weights_dim_2), ctypes.c_void_p(weights.ctypes.data), ctypes.byref(ierr_dim_1), ctypes.c_void_p(ierr.ctypes.data), ctypes.byref(interp_in_present), ctypes.byref(interp_in_dim_1), ctypes.byref(interp_in_dim_2), ctypes.c_void_p(interp_in.ctypes.data), ctypes.byref(interp_out_present), ctypes.byref(interp_out_dim_1), ctypes.byref(interp_out_dim_2), ctypes.c_void_p(interp_out.ctypes.data), ctypes.byref(eps_present), ctypes.byref(eps), ctypes.byref(extrap_present), ctypes.byref(extrap), ctypes.byref(rnorm_present), ctypes.byref(rnorm_dim_1), ctypes.c_void_p(rnorm.ctypes.data), ctypes.byref(ibudget_present), ctypes.byref(ibudget), ctypes.byref(chain_present), ctypes.byref(chain), ctypes.byref(pmode_present), ctypes.byref(pmode))

    # Return final results, 'INTENT(OUT)' arguments only.
    return pts, q, simps, weights, ierr, (interp_out if interp_out_present else None), (rnorm if rnorm_present else None)


class real_precision:
    ''''''

    # Declare 'r8'
    def get_r8(self):
        r8 = ctypes.c_int()
        clib.real_precision_get_r8(ctypes.byref(r8))
        return r8.value
    def set_r8(self, r8):
        raise(NotImplementedError('Module attributes with PARAMETER status cannot be set.'))
    r8 = property(get_r8, set_r8)

real_precision = real_precision()


class delsparse_mod:
    '''! This module contains the REAL_PRECISION R8 data type for 64-bit arithmetic
! and interface blocks for the DELAUNAYSPARSES and DELAUNAYSPARSEP
! subroutines for computing the Delaunay simplices containing interpolation
! points Q in R^D given data points PTS.'''

delsparse_mod = delsparse_mod()




import numpy as np
import fmodpy

# Matrix operations module.
rand = fmodpy.fimport("../axy_random.f90").random
print()


# Make sure that the random integer generation has desired behavior.
def _test_random_integer():
    print("Random integer..", flush=True)
    counts = {}
    trials = 1000000
    bins = 3
    for i in range(trials):
        val = rand.random_integer(max_value=bins)
        counts[val] = counts.get(val,0) + 1
    total_count = sum(counts.values())
    for (v,c) in sorted(counts.items()):
        ratio = (c / total_count)
        error = abs(ratio - 1/bins)
        assert (error < 0.001), f"Bad ratio of random integers had value {v} when generating with max_value = {bins}.\n  Ratio was {ratio}\n  Expected {1/bins}"
    print("  passed.", flush=True)

_test_random_integer()


# --------------------------------------------------------------------
#                        RANDOM_UNIT_VECTORS

# TODO: Add test with large number of vectors, ensure no Inf or Nan generated.
def _test_random_unit_vectors():
    print("Random unit vectors..", flush=True)
    # Generate test data.
    n = 5000 # number of points
    for d in (2,): # 3, 4, 5, 6, 7, 8, 9, 10, 15, 20, 25, 30, 60, 120):
        bd = 2 # full number of outputs
        br = 1 # reduced rank approximation

        # Random unit vectors.
        a = np.zeros((n,d), dtype="float32")
        rand.random_unit_vectors(a.T)

        # Overwrite a with perfectly spaced version for 2 dimensions.
        if (d == 2):
            _ = np.linspace(0, 2*np.pi, n)
            a[:,0] = np.cos(_)
            a[:,1] = np.sin(_)

        # Assert that the distance from the origin for each point is one.
        lengths = np.linalg.norm(a, axis=1)
        length_diff = abs(lengths - 1.0)
        index = np.argmax(length_diff)
        tolerance = 0.0001
        assert (max(length_diff) < tolerance), \
            "One of the vector lengths was significantly different than 1.\n" \
            f"  index = {index}\n" \
            f"  length = {lengths[index]}\n" \
            f"  tolerance = {tolerance}"

        # Assert that the distribution of distances between pairs is not
        #  significantly different than the expected analytic solution for
        #  distances between all pairs of points on a sphere.
        distances = np.concatenate([
            np.linalg.norm(a[:i,:] - a[i], axis=1)
            for i in range(1,len(a))
        ] + [
            np.linalg.norm(a[i+1:,:] - a[i], axis=1)
            for i in range(len(a)-1)
        ], axis=0)


        # If either fail, then plot the vectors.
        from tlux.math import project
        from tlux.plot import Plot, multiplot
        # Plot the actual points (in 3d).
        title = f"3D projection of {d}D points    |    CDF of distances between pairs of {d}D points"
        p1 = Plot(title)
        vecs = project(a, 3)
        if (n <= 100):
            for i in range(n):
                l = np.vstack([[0]*3, vecs[i]])
                p1.add(str(i), *l.T, mode="lines", show_in_legend=False)
        else:
            p1.add("vecs", *vecs.T, marker_size=2, color=1)
        # Plot the distribution of distances.
        p2 = Plot(title, "Distance between pair of points", "P[||p1-p2||] <= Distance")

        # Create the function that converts x and y positions into distances.
        radians = np.linspace(np.pi, 0, 1000)
        x_values = 1 + np.cos(radians)  # [0, ..., 2]
        y_values = np.sin(radians)  # [0, ..., 1, ..., 0]
        dist_values = np.linalg.norm(np.vstack((x_values, y_values)), axis=0)  # [0, ..., 2]

        # def cdf(dist):
        #     # Infer the "x" by pushing the value from the inverse function for distances.
        #     x = np.interp(dist, xp=dist_values, fp=x_values)
        #     # Map the x ∈ [0, 2] to the density.
        #     if (x <= 1): # x ** (1/d) 
        #         return ((1 - (1-x)**(1/(d-1))) / 2)
        #     else:
        #         return ((1/2) + ((x-1) ** (1/(d-1))) / 2)
        # p2.add_func("truth_guess", cdf, [0, 2], color=0, dash="dot")

        p2.add("coordinates", x_values, y_values, mode="lines", color=2)
        p2.add("dist(x)", x_values, dist_values, mode="lines", color=3)
        p2.add("xdist(d)", dist_values, x_values, mode="lines", color=4)

        # Add the distribution of distances.
        cdf_y = np.linspace(0, 1, 1000)
        cdf_x = np.percentile(distances, 100 * cdf_y)
        p2.add("distances CDF", cdf_x, cdf_y, mode="lines", color=1, show_in_legend=False)
        # Show both plots.
        multiplot([p1, p2], show_legend=False)

_test_random_unit_vectors()




import numpy as np


LARGE_TEST = False
COMPARE_AGAINST_SKLEARN = False
TEST_TREE = True
TEST_PRUNE = False
TEST_SORT = False


# × SUM(SQ) - 3.17 2.37 2.36 2.33 2.33
# ✓ OMP     - 2.51 2.31 2.30 2.26 2.27
# × PRE-ADD - 2.35 2.33 2.31 2.27 2.35


if COMPARE_AGAINST_SKLEARN:
    print()
    print("="*70)

    from tlux.system import Timer
    t = Timer()

    if LARGE_TEST: train, dim = 10000000, 10
    else:          train, dim = 5, 2
    test = 2
    leaf_size = 32
    k = 5
    print("Initializing data..", flush=True)
    np.random.seed(0)
    x = np.random.random(size=(train,dim)).astype("float32")
    z = np.random.random(size=(test,dim)).astype("float32")  # Use random z.
    # z = x[np.random.randint(0,len(x),size=(test,))]  # Use z points from x.
    print()
    print("x:", x.shape)
    print("z:", z.shape)
    # ----------------------------------------------------------------
    from tlux.approximate.balltree import BallTree as BT
    print()
    print("Fortran Ball Tree")
    t.start()
    # tree = BT(x, leaf_size=leaf_size)
    tree = BT(x[:len(x)//2], leaf_size=leaf_size, build=True) # build=False
    tree.add(x[len(x)//2 : (3*len(x))//4])
    tree.add(x[(3*len(x))//4 :])
    # tree.build()
    ct = t.stop()
    print("Construction time:", ct)
    t.start()
    d, i = tree.query(z, k=k)
    qt = t.stop()
    print("Query time:       ", qt)
    print("d: ",d[0])
    print("i: ",i[0])
    d1, i1 = d[0].copy(), i[0].copy()
    # ----------------------------------------------------------------
    from sklearn.neighbors import BallTree
    print()
    print("Sklearn Ball Tree")
    t.start()
    tree = BallTree(x, leaf_size=leaf_size)
    ct = t.stop()
    print("Construction time:", ct)
    t.start()
    d, i = tree.query(z, k=k)
    qt = t.stop()
    print("Query time:       ", qt)
    print("d: ",d[0])
    print("i: ",i[0])
    d2, i2 = d[0].copy(), i[0].copy()
    # ----------------------------------------------------------------
    print()
    print("Brute Force")
    # Convert to float64 (for regular test this does nothing, for integer...
    t.start()
    d = np.sqrt(np.sum(x**2, axis=1) + np.sum(z[0]**2) - 2 * np.dot(x, z[0]))
    i = np.argsort(d)
    qt = t.stop()
    print("Query time:", qt)
    i = i[:k]
    d = d[i]
    print("d: ",d)
    print("i: ",i)
    d3, i3 = d.copy(), i.copy()
    # ----------------------------------------------------------------
    max_diff = max(max(abs(d1-d2)), max(abs(d1-d3)))
    ds_match = max_diff < 2**(-13) # 2**(-26)
    is_match = np.all(i1 == i2) and np.all(i1 == i3)
    print()
    print(f"Max difference in distance calculations:\n   {max_diff:.3e}")
    assert (ds_match and is_match), f"\nERROR\n  is_match: {is_match}\n  ds_match: {ds_match} {max(abs(d1-d3)):.3e} {max(abs(d1-d2)):.3e}"


if TEST_TREE:
    from tlux.approximate.balltree import BallTree 
    np.random.seed(0)

    # if LARGE_TEST: size = (20000000,10)
    # if LARGE_TEST: size = (100000,1000)
    if LARGE_TEST: size = (50000000, 32)
    else:          size = (4, 2)
    print()
    print(f"Allocating array.. {size}", flush=True)

    x = np.random.random(size=size).astype("float32")

    # Handle printout based on test size.
    if len(x) < 20:
        print()
        print("x:",x.T.shape,"\n",x)
        print()

    print("Building tree..", flush=True)
    from tlux.system import Timer
    t = Timer()
    t.start()

    k = 3
    leaf_size = 1
    tree = BallTree(x, leaf_size=leaf_size, reorder=True)

    t.stop()
    print(f"done in {t()} seconds.", flush=True)

    if len(x) < 20:
        print()
        print('-'*70)
        print("Tree:")
        print(tree.tree.T)
        print()
        print("Usage:")
        print(tree.usage)
        print()
        print("Order, Radius:")
        print(tree.order)
        print(tree.radii)
        print('-'*70)

    z = np.random.random(size=(3,size[1]))

    if not LARGE_TEST: print("\nz: ",z,"\n")

    
    t.start()
    d,i = tree.query(z, k=k)
    t.stop()
    print("ball tree query in",t(),"seconds")
    if len(x) < 20: print(" usage:", tree.usage)
    d,i = d[0], i[0]
    # Do an approximate query.
    t.start()
    _,_ = tree.query(z, k=k, budget=100)
    t.stop()
    print("ball tree constrained query in",t(),"seconds")
    if len(x) < 20: print(" usage:", tree.usage)
    # Measure the "true" distances between points.
    t.start()
    true_dists = np.linalg.norm(x - z[0,:], axis=1)
    t.stop()
    print("linear search in",t(),"seconds")
    print()
    print("Tree/Truth")
    print("i: \n",i,"\n",np.argsort(true_dists)[:k])
    print("d: \n",d,"\n",np.sort(true_dists)[:k])


if TEST_PRUNE:
    from tlux.approximate.balltree import BallTree, prune
    
    tree_size = 14
    level = 3
    indices = np.zeros(2**level, order='F', dtype=np.int64)
    indices, found = prune.level(tree_size, level, indices)
    assert(found == 7)
    assert(tuple(indices) == (4,5,7,8,11,12,14,0))


    # Function for verifying correctness.
    def small_diff(v1,v2): return abs(v1 - v2) < .01
    # Build a tree over random points in 1 dimension.
    np.random.seed(3)
    pts = np.random.random(size=(100000,1))
    tree = BallTree()
    tree.add(pts)
    tree.build()

    vals = sorted((tree[0], tree[2], tree[1],
                  tree[(tree.tree.shape[1] + 2) // 2 + 1],
                  tree[(tree.tree.shape[1] + 2) // 2]
    ))
    print()
    for v in vals: print(v)
    print()
    assert(all(small_diff(v1,v2) for (v1,v2) in zip(
        vals, np.linspace(0,1,5))))

    tree.prune(4, method="inner")
    print(tree)
    print(tree[:len(tree)][0,:])
    print(sorted(tree[:len(tree)][0,:]))
    print()
    assert(all(small_diff(v1,v2) for (v1,v2) in zip(
        sorted(tree[:len(tree)][0,:]), np.linspace(0,1,len(tree)))))

    tree.prune(3, method="inner")
    print(tree)
    print(tree[:len(tree)][0,:])
    print(sorted(tree[:len(tree)][0,:]))
    print()
    assert(all(small_diff(v1,v2) for (v1,v2) in zip(
        sorted(tree[:len(tree)][0,:]), np.linspace(0,1,len(tree)))))

    tree.prune(2, method="inner")
    print(tree)
    print(tree[:len(tree)][0,:])
    print(sorted(tree[:len(tree)][0,:]))
    print()
    assert(all(small_diff(v1,v2) for (v1,v2) in zip(
        sorted(tree[:len(tree)][0,:]), np.linspace(0,1,len(tree)))))

    tree.prune(1, method="inner")
    print(tree)
    print(tree[:len(tree)][0,:])
    print(sorted(tree[:len(tree)][0,:]))
    print()
    assert(small_diff(tree[0][0], .5))


if TEST_SORT:
    if LARGE_TEST: N = 1000000
    else:          N = 10
    from tlux.approximate.balltree import fast_sort
    # Import a timer.
    from tlux.system import Timer
    t = Timer()
    # Generate test numbers.
    print()
    print(f"Generating {N} numbers..", flush=True)
    x = np.random.random(size=N)
    i = np.arange(len(x)) + 1
    print()
    # Test the fortran code.
    pts = x.copy()
    ids = i.copy()
    t.start()
    pts, ids = fast_sort.argsort(pts, ids)
    t.stop()
    # Check for correctness.
    ids_match = np.all(x[ids-1] == pts)
    is_sorted = np.all(np.diff(pts)>=0)
    try: assert(ids_match and is_sorted)
    except:
        print("ERROR")
        print(" ids_match: ",ids_match)
        print(" is_sorted: ",is_sorted)
        print()
        print("pts:", pts)
        print("ids:", ids)
        print()
        print(x)
        print(x[ids-1])
        exit()
    print("argsort: %.6f"%(t.total))
    # Test the NumPy code.
    pts = x.copy()
    ids = i.copy()
    t.start()
    ids = pts.argsort()
    t.stop()
    print("numpy:   %.6f"%(t.total))




# Fraction class that provides infinite-precision rational numbers
# (making use of Python default arbitrarily large integers).

from math import gcd
class UsageError(Exception): pass

# This class-operator wrapper automatially returns a `float` type if
# the operand is neither the same type as the operator nor an `int`.
def float_fallback(operator):
    def wrapped_operator(primary, other):
        # Call the standard operator if the types match.
        if (type(other) == type(primary)):
            return operator(primary, other)
        # Handle integers in a special way (so that users can put ints
        # in code without breaking fractions).
        elif (type(other) == int):
            return operator(primary, type(primary)(other))
        # Otherwise, cast "other" to the type of "primary", call the
        # operator, then cast the result back to a 'float' type.
        else:
            return float(operator(primary, type(primary)(other)))
    return wrapped_operator

# This class-operator wrapper automatially returns a `float` type if
# the operand is neither the same type as the operator nor an `int`.
def auto_same_type(operator):
    def wrapped_operator(primary, other):
        # Call the standard operator if the types match.
        if (type(other) == type(primary)):
            return operator(primary, other)
        # Otherwise, cast "other" into the same type.
        else: return operator(primary, type(primary)(other))
    return wrapped_operator

# This can be 'float_fallback' or 'auto_same_type'.
TYPE_HANDLER = auto_same_type

# A fraction instance. Initialized as:
# 
#     Fraction(numerator=0, denominator=1)
# 
# The arguments to the Fraction must either be Python integers or
# convertible to Python float via the 'float' built-in function.
class Fraction:
    def __init__(self, numerator=0, denominator=1, reduce=True):
        # If "None" was provided, then set this fraction to have "None" values.
        if ((numerator is None) or (denominator is None)):
            self._numerator = None
            self._denominator = None
            return
        # If a Fraction was provided, shortcut and exit.
        elif (type(numerator) == Fraction):
            self._numerator   = numerator.numerator
            self._denominator = numerator.denominator
            return
        # Convert input arguments to the appropriate type.
        if (type(numerator) != int):
            try:    numerator = float(numerator)
            except: raise(UsageError(f"Unsupported numerator number type '{type(numerator)}', could not convert to float."))
        if (type(denominator) != int):
            try:    denominator = float(denominator)
            except: raise(UsageError(f"Unsupported denominator number type '{type(denominator)}', could not convert to float."))
        # Convert non-integer numerator and denominator appropriately.
        if (type(numerator) == float):
            # Handle 'infinity' in an expected way (without errors).
            if (abs(numerator) == float('inf')):
                if (abs(denominator) != float('inf')):
                    v = int(numerator < 0)
                    numerator, denominator = (-1)**(numerator < 0), 0
                elif (denominator < 0):
                    numerator, denominator = -numerator, -denominator
            # The normal usage case.
            else:
                numerator, _ = numerator.as_integer_ratio()
                denominator *= _
        if (type(denominator) == float):
            # Handle 'infinity' in an expected way (without errors).
            if (abs(denominator) == float('inf')):
                if (abs(numerator) != float('inf')):
                    numerator, denominator = 0, 1
            # The normal usage case.
            else:
                denominator, _ = denominator.as_integer_ratio()
                numerator *= _
        # Check to make sure the "infinity" cases are handled consistently.
        if (numerator == 0):
            if (denominator != 0): denominator //= abs(denominator)
        elif (denominator == 0):   numerator   //= abs(numerator)
        # Transfer the negativity to the numerator if possible.
        if   (numerator < 0) and (denominator < 0):
            numerator, denominator = -numerator, -denominator
        elif (numerator >= 0) and (denominator < 0):
            numerator, denominator = -numerator, -denominator
        # If this is indeterminate form, do not reduce!
        if (numerator == 0 == denominator): reduce = False
        # Simplify the numerator and denominator if appropriate.
        if reduce:
            divisor = gcd(numerator, denominator)
            numerator   //= divisor
            denominator //= divisor
        # Store the numerator and denominator of this fraction.
        self._numerator = numerator
        self._denominator = denominator

    # Provide access to the numerator and denominator of this Fraction.
    @property
    def numerator(self): return self._numerator
    @property
    def denominator(self): return self._denominator

    # Convert this fraction into a float.
    def __float__(self):
        if (self.denominator == 0):
            if   (self.numerator > 0): return float('inf')
            elif (self.numerator < 0): return -float('inf')
            else:                      return float(0)
        else: return (self.numerator / self.denominator)
    # Convert this fraction to an integer.
    def __int__(self):   return (self.numerator // self.denominator)

    # Provide a represenation of this Fraction.
    def __repr__(self): return f"{self.__class__.__name__}({self.numerator}, {self.denominator})"
    # Provide a string representation of this Fraction.
    def __str__(self): return f"{self.numerator} / {self.denominator}"

    # Add two fractions.
    @TYPE_HANDLER
    def __add__(a, b):
        an, bn = a.numerator, b.numerator
        ad, bd = a.denominator, b.denominator
        return Fraction( an * bd + bn * ad,  ad * bd )
    @TYPE_HANDLER
    def __radd__(b, a): return a + b

    #  Subtract two fractions.
    @TYPE_HANDLER
    def __sub__(a, b):
        an, bn = a.numerator, b.numerator
        ad, bd = a.denominator, b.denominator
        return Fraction( an * bd - bn * ad,  ad * bd )
    @TYPE_HANDLER
    def __rsub__(a, b): return -a + b

    # Multiply two fractions.
    @TYPE_HANDLER
    def __mul__(a, b):
        return Fraction(a.numerator * b.numerator, a.denominator * b.denominator)
    @TYPE_HANDLER
    def __rmul__(a, b): return a * b

    # Divide two fractions.
    @TYPE_HANDLER
    def __truediv__(a, b):
        return Fraction(a.numerator * b.denominator, a.denominator * b.numerator)
    @TYPE_HANDLER
    def __rtruediv__(b, a):
        return a / b

    # Divide two fractions (with integer result).
    def __floordiv__(a, b):
        b = Fraction(b)
        result = a / b
        result.numerator -= result.numerator % result.denominator
        return result.numerator // result.denominator
    def __rfloordiv__(b, a):
        a = Fraction(a)
        return a // b

    # Compute a mod b (and b mod a).
    @TYPE_HANDLER
    def __mod__(a, b):
        mult = a // b
        return a - b * mult
    @TYPE_HANDLER
    def __rmod__(b, a):
        return a % b

    # Compute a raised to the power b.
    # 
    # WARNING: Precision is lost if a root is taken (fractional power).
    def __pow__(a, b):
        # Handle the integer power case.
        if (type(b) == int):
            if (b == 0):  return Fraction(1,1,reduce=False)
            elif (b > 0): return Fraction(a.numerator**b, a.denominator**b)
            else:         return Fraction(a.denominator**(abs(b)), a.numerator**(abs(b)))
        else:
            # Handle the fractional power case (break up into two steps).
            b = Fraction(b)
            # Handle infinite values in a reasonable way.
            if (b.denominator == 0):
                if   (b.numerator > 0): return Fraction(a * float('inf'))
                elif (b.numerator < 0): return Fraction(0)
                else: b = Fraction(0,1)
            # If this is essentially an integer, just use that operation.
            elif (b.denominator == 1): return a ** b.numerator
            # First do the integer power (exact operation).
            intermediate = a ** abs(b.numerator)
            # Second do the inexact operation (converts to a float).
            result = float(intermediate) ** (1 / b.denominator)
            if (b.numerator < 0):
                if (result != 0): return 1 / result
                else:             return Fraction(1,0)
            return result

    # When taking a Fraction as a power..
    def __rpow__(b, a):
        return Fraction(a) ** b

    # Negate this Fraction.
    def __neg__(a): return Fraction(-a.numerator, a.denominator, reduce=False)

    # Make this Fraction positive.
    def __abs__(a): return Fraction(abs(a.numerator), a.denominator, reduce=False)

    # a equals b. (special case is False for comparison with None)
    def __eq__(a, b):
        b = Fraction(b)
        return (a.numerator == b.numerator) and (a.denominator == b.denominator)

    # a less than b.
    def __lt__(a, b):
        b = Fraction(b)
        return a.numerator * b.denominator < a.denominator * b.numerator

    # a greater than b
    def __gt__(a, b):
        b = Fraction(b)
        return a.numerator * b.denominator > a.denominator * b.numerator

    # a less than or equal to b
    def __le__(a, b):
        b = Fraction(b)
        return a.numerator * b.denominator <= a.denominator * b.numerator

    # a greater than or equal to b
    def __ge__(a, b):
        b = Fraction(b)
        return a.numerator * b.denominator >= a.denominator * b.numerator

    # a not equal to zero
    def __bool__(a):
        return (a.numerator != 0) and (a.denominator != 0)

    # Return the rounded version of this Fraction.
    def __round__(self, *args, **kwargs):
        return round(float(self), *args, **kwargs)


# Some minor testing and demonstration code. Anything not tested here
# has not been explicitly tested!! Test cases will be built as more
# features are used.
def _test_Fraction(display=False, print=lambda *args, **kwargs: None):
    # If display is desired, re-assign the real print function.
    if display:
        import builtins
        print = builtins.print

    print()
    print("1/3              : ",1/3)
    print("a = Fraction(1,3): ",Fraction(1,3))
    print("b = Fraction(1/3): ",Fraction(1/3))

    a = Fraction(1,3)
    assert(a.numerator == 1)
    assert(a.denominator == 3)
    b = Fraction(1/3)
    assert(a != b)

    print()
    print("a - b:             ",a - b)
    assert((a - b) == Fraction(1, 54043195528445952))
    print("float(a - b):      ",float(a - b))
    print("a_sqrt = a**(1/2): ",a**(1/2))
    a_sqrt = a**(1/2)
    print()

    print("error = a - a_sqrt**2: ",a - a_sqrt**2)
    error = a - a_sqrt**2
    print("float(error):          ",float(error))
    print("error - float(a-b):    ",float(error - float(a-b)))
    # Error should certainly be less than SQRT error for rounded fractions.
    assert(error <= float(a-b)**(1/2))  
    print()

    print("Fraction():       ",Fraction())
    assert(Fraction() == Fraction(0,1))
    print("bool(Fraction()): ",bool(Fraction()))
    assert(not bool(Fraction()))
    print("bool(a):          ",bool(a))
    assert(bool(a))
    print()

    print("c = Fraction(float('inf')): ",Fraction(float('inf')))
    c = Fraction(float('inf'))
    assert(c == Fraction(1,0))
    print("Fraction(-float('inf')):    ",Fraction(-float('inf')))
    assert(Fraction(-float('inf')) == Fraction(-1,0))
    print("10 / c:    ",Fraction(10) / c)
    assert((Fraction(10) / c) == Fraction(0,1.))
    print("c / 10:    ",c / Fraction(10))
    assert((c / Fraction(10)) == Fraction(1.,0))
    print("c * 100.0: ",c * 100.0)
    assert((c * 100.0) == float('inf'))
    print("-c * 10:   ",-c)
    assert((-c*10) == Fraction(-1,0))
    print("c - c:     ",c - c)
    assert((c-c) == Fraction(0,0))
    print("float(c-c) ", float(c-c))
    assert(float(c-c) == 0.0)
    print()
    print("a * (c - c)", a*(c - c))
    assert((a*(c-c)) == Fraction(0,0))
    print("a / (c - c)", a/(c - c))
    assert((a/(c-c)) == Fraction(0,0))
    print("a ** c     ", a**c)
    assert((a**c) == Fraction(1,0))
    print("a ** (-c)  ", a**(-c))
    assert((a**(-c)) == Fraction(0,1))
    print("(-a) ** c  ", (-a)**c)
    assert(((-a)**c) == Fraction(-1,0))
    print()

    print("d = Fraction(1,float('inf')): ",Fraction(1,float('inf')))
    d = Fraction(1,float('inf'))
    assert(d == Fraction(0,1))
    print("Fraction(1,-float('inf')):    ",Fraction(1,-float('inf')))
    assert(Fraction(1,-float('inf')) == Fraction(0,1))
    print("10 / d:  ",10 / d)
    assert((10/d) == Fraction(1,0))
    print("d / 10:  ",d / 10)
    assert((d/10) == Fraction(0,1))
    print("d * 100: ",d * 100)
    assert((d * 100) == Fraction(0,1))
    print("c * d:   ",c * d)
    assert((c * d) == Fraction(0,0))
    print("-d:      ",-d)
    assert((-d) == Fraction(0,1))
    print()

    # Make sure that the power operator behaves in expected ways.
    a = Fraction(7,3)
    assert(a**2 == a**2.)
    assert(a**(-2) == a**(-2.))
    assert(Fraction(2.0, -1/2) == Fraction(-4))


if __name__ == "__main__":
    print()
    print("Testing Fraction..")
    _test_Fraction(display=False)
    print("done.")
    



import os, sys, time
import numpy as np
import fmodpy
from tlux.random import well_spaced_ball as random
from tlux.random import ball as random
from memory_profiler import profile

# Get the matrix operations code from the main directory.
source_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
source_path = os.path.join(source_dir, "matrix_operations.f90")
mops = fmodpy.fimport(source_path, blas=True, lapack=True, omp=True).matrix_operations
# Load an old copy of matrix operations that has been used heavily.
old_mops = fmodpy.fimport("old_matrix_operations.f90", blas=True, lapack=True, omp=True).matrix_operations

print("Initializing data..")
n = 10000000
d = 64
x = np.asarray(random(n, d).T, dtype="float32", order="F")
shift = np.zeros((d,), dtype="float32")
vecs = np.zeros((d,d), dtype="float32", order="F")

print("Calling radialize..")
def run():
    start = time.time()    
    _new = mops.radialize(x=np.array(x, order="F"), shift=shift.copy(), vecs=np.array(vecs, order="F"), max_to_square=n)
    total = time.time() - start
    start = time.time()    
    _old = old_mops.radialize(x=np.array(x, order="F"), shift=shift.copy(), vecs=np.array(vecs, order="F"))
    old_total = time.time() - start
    for a,b in zip(_old, _new):
        diff = np.abs(a-b).max()
        assert diff < 2**(-14), f"Difference between implementations ({diff:.8f}) is too large."
    print(f"Comparison passed, no issues.. ({old_total/total:.1f}x speedup)")
    return total
print(f"Done in {run():.2f} seconds.")



# -----------------------------------------------------------------------
# License: MIT
#  Author: Thomas C.H. Lux
#  Source: https://github.com/tchlux/tlux
#    Date: 2022/04
# 
# Description:
#   A class that implements a numpy-only evaluation mode for AXY models.
# 
# -----------------------------------------------------------------------


import numpy as np


# Build a class that contains pointers to the model internals, allowing
#  python attribute access to all of the different components of the models.
#  Define a "predict" method that evaluates the model using only numpy
#  the same way as would be done in the compiled code.
class AxyNumpy:
    # Static method for loading a saved model from a path.
    def load(path):
        # Read the file.
        import json
        with open(path, "r") as f:
            attrs = json.loads(f.read())
        # Load the attributes of the model.
        for key in attrs:
            value = attrs[key]
            if (key[-4:] == "_map"):
                value = [np.asarray(l) for l in value]
            elif (type(value) is list): 
                value = np.asarray(value, dtype="float32")
            attrs[key] = value
        # Return an initialized model.
        return AxyNumpy(config=attrs.pop("config",None), model=attrs.pop("model", None), **attrs)


    # Given a configuration object and a model array, store internal parameters.
    def __init__(self, config=None, model=None, **attrs):
        self.config = config
        self.model = model
        for attr in attrs: setattr(self, attr, attrs[attr])
        # Store the attributes of the model in self by breaking out
        #  the individual components 
        if (self.config is not None) and (self.model is not None):
            self.a_embeddings  = self.model[self.asev-1:self.aeev].reshape(self.ade, self.ane, order="F")
            self.a_input_vecs  = self.model[self.asiv-1:self.aeiv].reshape(self.adi, self.ads, order="F")
            self.a_input_shift = self.model[self.asis-1:self.aeis].reshape(self.ads, order="F")
            self.a_state_vecs  = self.model[self.assv-1:self.aesv].reshape(self.ads, self.ads, max(0,self.ans-1), order="F")
            self.a_state_shift = self.model[self.asss-1:self.aess].reshape(self.ads, max(0,self.ans-1), order="F")
            self.a_output_vecs = self.model[self.asov-1:self.aeov].reshape(self.adso, self.ado, order="F")
            self.m_embeddings  = self.model[self.msev-1:self.meev].reshape(self.mde, self.mne, order="F")
            self.m_input_vecs  = self.model[self.msiv-1:self.meiv].reshape(self.mdi, self.mds, order="F")
            self.m_input_shift = self.model[self.msis-1:self.meis].reshape(self.mds, order="F")
            self.m_state_vecs  = self.model[self.mssv-1:self.mesv].reshape(self.mds, self.mds, max(0,self.mns-1), order="F")
            self.m_state_shift = self.model[self.msss-1:self.mess].reshape(self.mds, max(0,self.mns-1), order="F")
            self.m_output_vecs = self.model[self.msov-1:self.meov].reshape(self.mdso, self.mdo, order="F")
            self.ax_shift = self.model[self.aiss-1:self.aise]
            self.ay_shift = self.model[self.aoss-1:self.aose]
            self.x_shift = self.model[self.miss-1:self.mise]
            self.y_shift = self.model[self.moss-1:self.mose]


    # Allow square brackets to access attributes of this model and its configuration.
    def __getitem__(self, attr):
        if (attr in self.config):
            return self.config[attr]
        else:
            return self.__getattribute__(attr)


    # Allow the "." operator to access attributes of this model and its configuration.
    def __getattr__(self, attr):
        if (attr in self.config):
            return self.config[attr]
        else:
            return self.__getattribute__(attr)


    # Create a summary string for this model.
    def __str__(self, vecs=False):
        # A function for creating a byte-size string from an integer.
        def _byte_str(byte_size):
            if (byte_size < 2**10):
                byte_size = f"{byte_size} bytes"
            elif (byte_size < 2**20):
                byte_size = f"{byte_size//2**10:.1f}KB"
            elif (byte_size < 2**30):
                byte_size = f"{byte_size//2**20:.1f}MB"
            elif (byte_size < 2**40):
                byte_size = f"{byte_size//2**30:.1f}GB"
            else:
                byte_size = f"{byte_size//2**40:.1f}TB"
            return byte_size
        # Calculate the byte size of this model (excluding python descriptors).
        # TODO: Not all configs are 4 bytes, do more expensive sum over actual sizes?
        byte_size = len(self.config)*4 + self.model.dtype.itemsize*self.model.size
        byte_size = _byte_str(byte_size)
        # Create a function that prints the actual contents of the arrays.
        if vecs: to_str = lambda arr: "\n    " + "\n    ".join(str(arr).split("\n")) + "\n"
        else:    to_str = lambda arr: "\n"
        # Provide details (and some values where possible).
        return (
            f"AXY model ({self.total_size} parameters) [{byte_size}]\n"+
            (" aggregator model\n"+
            f"  input dimension  {self.adn}\n"+
            f"  output dimension {self.ado}\n"+
            f"  state dimension  {self.ads}\n"+
            f"  number of states {self.ans}\n"+
           (f"  embedding dimension  {self.ade}\n"+
            f"  number of embeddings {self.ane}\n"
             if self.ane > 0 else "")+
            f"  embeddings   {self.a_embeddings.shape}  "+to_str(self.a_embeddings)+
            f"  input vecs   {self.a_input_vecs.shape}  "+to_str(self.a_input_vecs)+
            f"  input shift  {self.a_input_shift.shape} "+to_str(self.a_input_shift)+
            f"  state vecs   {self.a_state_vecs.shape}  "+to_str(self.a_state_vecs)+
            f"  state shift  {self.a_state_shift.shape} "+to_str(self.a_state_shift)+
            f"  output vecs  {self.a_output_vecs.shape} "+to_str(self.a_output_vecs)+
             "\n" if (self.a_output_vecs.size > 0) else "") +
            (" model\n"+
            f"  input dimension  {self.mdn}\n"+
            f"  output dimension {self.mdo}\n"+
            f"  state dimension  {self.mds}\n"+
            f"  number of states {self.mns}\n"+
           (f"  embedding dimension  {self.mde}\n"+
            f"  number of embeddings {self.mne}\n"
             if self.mne > 0 else "")+
            f"  embeddings   {self.m_embeddings.shape}  "+to_str(self.m_embeddings)+
            f"  input vecs   {self.m_input_vecs.shape}  "+to_str(self.m_input_vecs)+
            f"  input shift  {self.m_input_shift.shape} "+to_str(self.m_input_shift)+
            f"  state vecs   {self.m_state_vecs.shape}  "+to_str(self.m_state_vecs)+
            f"  state shift  {self.m_state_shift.shape} "+to_str(self.m_state_shift)+
            f"  output vecs  {self.m_output_vecs.shape} "+to_str(self.m_output_vecs)
             if (self.m_output_vecs.size > 0) else "")
        )


    # Given a categorical input array, construct a dictionary for
    #  mapping the unique values in the columns of the array to integers.
    def _i_map(self, xi):
        if (len(xi.dtype) > 0):
            xi_map = [np.unique(xi[n]) for n in xi.dtype.names]
        else:
            xi_map = [np.unique(xi[:,i]) for i in range(xi.shape[1])]
        return xi_map


    # Given a categorical input array (either 2D or struct), map this
    #  array to an integer encoding matrix with the same number of
    #  columns, but unique integers assigned to each unique value.
    # 
    # TODO: This is very slow and doesn't use any parallelism right now,
    #       could be much faster for large data by paralellizing.
    def _i_encode(self, xi, xi_map):
        xi_rows = xi.shape[0]
        xi_cols = len(xi.dtype) or xi.shape[1]
        _xi = np.zeros((xi_rows, xi_cols), dtype="int32", order="C")
        base = 1
        for i in range(xi_cols):
            # Construct the integer map, retrieve the column of unique values.
            # The integer map should not have any numbers shared across yi columns!
            int_map = {xi_map[i][j]:base+j for j in range(len(xi_map[i]))}
            base += len(xi_map[i])
            vals = (xi[:,i:i+1] if len(xi.dtype) == 0 else xi[xi.dtype.names[i]])
            # Assign all the integers.
            for j in range(xi_rows):
                _xi[j,i] = int_map.get(vals[j,0], 0)
        return _xi


    # Convert all inputs to the AXY model into the expected numpy format.
    def _to_array(self, ax, axi, sizes, x, xi, y, yi):
        # Get the number of inputs.
        if   (y  is not None): nm = len(y)
        elif (yi is not None): nm = len(yi)
        elif (x  is not None): nm = len(x)
        elif (xi is not None): nm = len(xi)
        elif (sizes is not None): nm = len(sizes)
        # Make sure that all inputs are numpy arrays.
        if (y is not None):  y = np.asarray(y, dtype="float32", order="C")
        else:                y = np.zeros((nm,0), dtype="float32", order="C") 
        if (yi is not None): yi = np.asarray(yi)
        else:                yi = np.zeros((nm,0), dtype="int32", order="C")
        if (x is not None): x = np.asarray(x, dtype="float32", order="C")
        else:               x = np.zeros((nm,0), dtype="float32", order="C")
        if (xi is not None): xi = np.asarray(xi)
        else:                xi = np.zeros((nm,0), dtype="int32", order="C")
        if (sizes is not None): sizes = np.asarray(sizes, dtype="int32")
        else:                   sizes = np.zeros(0, dtype="int32")
        na = sizes.sum()
        if (ax is not None): ax = np.asarray(ax, dtype="float32", order="C")
        else:                ax = np.zeros((na,0), dtype="float32", order="C")
        if (axi is not None): axi = np.asarray(axi)
        else:                 axi = np.zeros((na,0), dtype="int32", order="C")
        # Make sure that all inputs have the expected shape.
        assert (len(y.shape) in {1,2}), f"Bad y shape {y.shape}, should be 1D or 2D matrix."
        assert (len(yi.shape) in {1,2}), f"Bad yi shape {yi.shape}, should be 1D or 2D matrix."
        assert (len(x.shape) in {1,2}), f"Bad x shape {x.shape}, should be 1D or 2D matrix."
        assert (len(xi.shape) in {1,2}), f"Bad xi shape {xi.shape}, should be 1D or 2D matrix."
        assert (len(ax.shape) in {1,2}), f"Bad ax shape {ax.shape}, should be 1D or 2D matrix."
        assert (len(axi.shape) in {1,2}), f"Bad axi shape {axi.shape}, should be 1D or 2D matrix."
        assert (len(sizes.shape) == 1), f"Bad sizes shape {sizes.shape}, should be 1D int vector."
        # Reshape inputs to all be two dimensional (except sizes).
        if (len(y.shape) == 1): y = y.reshape((-1,1))
        if (len(yi.shape) == 1) and (len(yi.dtype) == 0): yi = yi.reshape((-1,1))
        if (len(x.shape) == 1): x = x.reshape((-1,1))
        if (len(xi.shape) == 1) and (len(xi.dtype) == 0): xi = xi.reshape((-1,1))
        if (len(ax.shape) == 1): ax = ax.reshape((-1,1))
        if ((len(axi.shape) == 1) and (len(axi.dtype) == 0)): axi = axi.reshape((-1,1))
        mdo = y.shape[1]
        mdn = x.shape[1]
        adn = ax.shape[1]
        # Handle mapping "xi" into integer encodings.
        xi_cols = len(xi.dtype) or xi.shape[1]
        if (xi_cols > 0):
            if (len(self.xi_map) == 0):
                self.xi_map = self._i_map(xi)
            else:
                assert (xi_cols == len(self.xi_map)), f"Bad number of columns in 'xi', {xi_cols}, expected {len(self.xi_map)} columns."
            xi = self._i_encode(xi, self.xi_map)
            mne = sum(map(len, self.xi_map))
        else: mne = 0
        # Handle mapping "axi" into integer encodings.
        axi_cols = len(axi.dtype) or axi.shape[1]
        if (axi_cols > 0):
            if (len(self.axi_map) == 0):
                self.axi_map = self._i_map(axi)
            else:
                assert (axi_cols == len(self.axi_map)), f"Bad number of columns in 'axi', {axi_cols}, expected {len(self.axi_map)} columns."
            axi = self._i_encode(axi, self.axi_map)
            ane = sum(map(len, self.axi_map))
        else: ane = 0
        # Handle mapping "yi" into integer encodings.
        yi_cols = len(yi.dtype) or yi.shape[1]
        if (yi_cols > 0):
            if (len(self.yi_map) == 0):
                self.yi_map = self._i_map(yi)
            else:
                assert (yi_cols == len(self.yi_map)), f"Bad number of columns in 'yi', {yi_cols}, expected {len(self.yi_map)} columns."
            yi = self._i_encode(yi, self.yi_map)
            yne = sum(map(len, self.yi_map))
        else: yne = 0
        # Handle mapping integer encoded "yi" into a single real valued y.
        if (yne > 0):
            embedded = np.concatenate((
                np.zeros((1,yne), dtype="float32"),
                np.identity(yne, dtype="float32")), axis=0)
            _y = np.zeros((nm, mdo+yne), dtype="float32")
            _y[:,:mdo] = y[:,:]
            for i in range(yi.shape[1]):
                _y[:,mdo:] += embedded[yi[:,i]]
            y = _y
            mdo += yne
        # Return all the shapes and numpy formatted inputs.
        return nm, na, mdn, mne, mdo, adn, ane, yne, y, x, xi, ax, axi, sizes


    # Calling this model is an alias for 'AXY.predict'.
    def __call__(self, *args, **kwargs):
        return self.predict(*args, **kwargs)


    # Make predictions for new data.
    def predict(self, x=None, xi=None, ax=None, axi=None, sizes=None):
        # Evaluate the model at all data.
        assert ((x is not None) or (xi is not None) or (sizes is not None)), "AXY.predict requires at least one of 'x', 'xi', or 'sizes' to not be None."
        # Make sure that 'sizes' were provided for aggregator (aggregate) inputs.
        if ((ax is not None) or (axi is not None)):
            assert (sizes is not None), "AXY.predict requires 'sizes' to be provided for aggregator input sets (ax and axi)."
        # Make sure that all inputs are numpy arrays.
        nm, na, mdn, mne, mdo, adn, ane, yne, _, x, xi, ax, axi, sizes = (
            self._to_array(ax, axi, sizes, x, xi, None, None)
        )
        # Embed the inputs into the purely positional form.
        ade = self.ade
        ads = self.ads
        ado = self.ado
        mde = self.mde
        mds = self.mds
        if (self.mdo != 0):
            mdo = self.mdo
        else:
            mdo = self.ado
        # Compute the true real-vector input dimensions given embeddings.
        adn += ade
        mdn += mde + ado
        # ------------------------------------------------------------
        # Initialize storage for all arrays needed at evaluation time.
        #   If there are integer embeddings, expand "ax" and "x" to have
        #   space to hold those embeddings.
        if (self.ade > 0):
            _ax = np.zeros((ax.shape[0],ax.shape[1]+self.ade), dtype="float32", order="C")
            _ax[:,:ax.shape[1]] = ax
            ax = _ax
        ay = np.zeros((na, ado), dtype="float32", order="F")
        if (self.mde > 0) or (self.ado > 0):
            _x = np.zeros((x.shape[0],self.mdi), dtype="float32", order="C")
            _x[:,:x.shape[1]] = x
            x = _x
        y = np.zeros((nm, mdo), dtype="float32", order="C")
        # ------------------------------------------------------------
        # Perform all data and array validation checks to make sure sizes are correct.
        assert (x.shape[0] == y.shape[0]), f"x {x.shape} and y {y.shape} do not have the same numer of rows."
        assert (x.shape[1] == self.mdi), f"x {x.shape} must have {self.mdi} columns."
        assert ((self.mdo == 0) or (self.mdo == y.shape[1])), f"y {y.shape} must have {self.mdo} columns."
        assert ((self.mdo > 0) or (self.ado == y.shape[1])), f"y {y.shape} must have {self.ado} columns."
        assert ((self.mne == 0) or (xi.shape[0] == x.shape[0])), f"xi {xi.shape} must have same number of rows as x {x.shape}."
        if (xi.size > 0):
            assert ((xi.min() >= 0) and (xi.max() <= self.mne)), f"xi [{xi.min()},{xi.max()}] must be in range [0,{self.mne}]."
        assert ((self.adi == 0) or (sizes.size == y.shape[0])), f"sizes {sizes.size} must have size equal to number of rows in y {y.shape}."
        assert (sizes.sum() == ax.shape[0]), f"sum(sizes) = {sizes.sum()} but should equal number of rows in ax {ax.shape}."
        assert (self.adi == ax.shape[1]), f"ax {ax.shape} must have {self.adi} columns."
        assert (ax.shape[0] == axi.shape[0]), f"ax {ax.shape} number of rows does not match axi {axi.shape}."
        if (axi.size > 0):
            assert ((axi.min() >= 0) and (axi.max() <= self.ane)), f"axi [{axi.min()},{axi.max()}] must be in range [0,{self.ane}]."
        # ------------------------------------------------------------
        # Perform the embedding operation, putting all embedding vectors into input format.
        if (self.ade > 0):
            a_known = axi > 0
            a_embeddings = (
                (self.a_embeddings[:,axi.flatten()-1].reshape((-1,)+axi.shape) # get all embedding vectors (ade, an, axi.shape[1])
                 * a_known) # zero out the "unkown" embeddings (integer value = 0)
                .sum(axis=-1) / np.maximum(1,a_known.sum(axis=-1)) # take mean of however many embeddings were known (count > 0)
            )
            ax[:, self.adn:self.adn+self.ade] = a_embeddings.T
        if (self.mde > 0):
            m_known = xi > 0
            m_embeddings = (
                (self.m_embeddings[:,xi.flatten()-1].reshape((-1,)+xi.shape) # get all embedding vectors (mde, mn, xi.shape[1])
                 * m_known) # zero out the "unkown" embeddings (integer value = 0)
                .sum(axis=-1) / np.maximum(1,m_known.sum(axis=-1)) # take mean of however many embeddings were known (count > 0)
            )
            x[:, self.mdn:self.mdn+self.mde] = m_embeddings.T
        # ------------------------------------------------------------
        # Evaluate aggregator model first, if it exists.
        if (self.adi > 0):
            state = ax
            state[:,:self.adn] += self.ax_shift[:]
            if (self.ans > 0):
                state = np.maximum(self.discontinuity, self.a_input_shift
                                   + np.matmul(state, self.a_input_vecs))
                for i in range(self.ans - 1):
                    state = np.maximum(self.discontinuity, self.a_state_shift[:,i]
                                       + np.matmul(state, self.a_state_vecs[:,:,i]))
            state = np.matmul(state, self.a_output_vecs)
            # Aggregate over set outputs.
            if (self.mdo > 0):
                ay_start = self.mdn + self.mde
                set_start = set_end = 0
                for i in range(sizes.shape[0]):
                    set_end += sizes[i]
                    x[i,ay_start:] = state[set_start:set_end].mean(axis=0)
                    set_start = set_end
                x[:,ay_start:] += self.ay_shift[:]
            else:
                set_start = set_end = 0
                for i in range(sizes.shape[0]):
                    set_end += sizes[i]
                    y[i,:] = state[set_start:set_end].mean(axis=0)
                    set_start = set_end
        # Evaluate the positional model second, if it exists.
        if (self.mdo > 0):
            state = x
            state[:,:self.mdn] += self.x_shift[:]
            if (self.mns > 0):
                state = np.maximum(self.discontinuity, self.m_input_shift
                                   + np.matmul(state, self.m_input_vecs))
                for i in range(self.mns - 1):
                    state = np.maximum(self.discontinuity, self.m_state_shift[:,i]
                                       + np.matmul(state, self.m_state_vecs[:,:,i]))
            y[:,:] = np.matmul(state, self.m_output_vecs)            
        # Add output shift.
        y[:,:] += self.y_shift[:]
        # ------------------------------------------------------------
        # If there are embedded y values in the output, return them to the format at training time.
        if (len(self.yi_map) > 0):
            yne = sum(self.yi_sizes)
            _y = [y[:,i] for i in range(y.shape[1]-yne)]
            for i in range(len(self.yi_map)):
                start = self.yi_starts[i]
                size = self.yi_sizes[i]
                _y.append(
                    self.yi_map[i][np.argmax(y[:,start:start+size], axis=1)]
                )
            return np.asarray(_y).T
        else:
            return y



# Ball Tree wrapper for Fortran module.

# TODO: Investigate the correctness of the partial tree traversal
#       (budget < size(order)) when there are duplicate data points.
# TODO: Make queries over trees with size > built do a brute-force
#       search over the unbuilt portion of the tree (when exact).

import os
import numpy as np

# Count the number of CPUs for setting default parallelism.
from multiprocessing import cpu_count

# Import the Fortran utilities.
try:
    from tlux.approximate.balltree.ball_tree import ball_tree
    from tlux.approximate.balltree.fast_sort import fast_sort
    from tlux.approximate.balltree.fast_select import fast_select
    from tlux.approximate.balltree.prune import prune
except:
    from tlux.setup import build_balltree
    ball_tree, fast_sort, fast_select, prune = build_balltree()

# ------------------------------------------------------------------
#                        FastSort method
# 
# This routine uses a combination of QuickSort (with modestly
# intelligent pivot selection) and Insertion Sort (for small arrays)
# to achieve very fast average case sort times for both random and
# partially sorted data. The pivot is selected for QuickSort as the
# median of the first, middle, and last values in the array.
# 
# Arguments:
# 
#   VALUES   --  A 1D array of real numbers.
# 
# Optional:
# 
#   INDICES  --  A 1D array of original indices for elements of VALUES.
#   MIN_SIZE --  An positive integer that represents the largest
#                sized VALUES for which a partition about a pivot
#                is used to reduce the size of a an unsorted array.
#                Any size less than this will result in the use of
#                INSERTION_ARGSORT instead of ARGPARTITION.
# 
# Output:
# 
#   The elements of the array VALUES are sorted and all elements of
#   INDICES are sorted symmetrically (given INDICES = 1, ...,
#   SIZE(VALUES) beforehand, final INDICES will show original index
#   of each element of VALUES before the sort operation).
# 
def argsort(values, indices=None, min_size=None):
    indices = np.arange(len(values))
    return fast_sort.argsort(values, indices, min_size=min_size)


# ------------------------------------------------------------------
#                       FastSelect method
# 
# Given VALUES list of numbers, rearrange the elements of VALUES
# such that the element at index K has rank K (holds its same
# location as if all of VALUES were sorted). Symmetrically rearrange
# array INDICES to keep track of prior indices.
# 
# This algorithm uses the same conceptual approach as Floyd-Rivest,
# but instead of standard-deviation based selection of bounds for
# recursion, a rank-based method is used to pick the subset of
# values that is searched. This simplifies the code and improves
# readability, while achieving the same tunable performance.
# 
# Arguments:
# 
#   VALUES   --  A 1D array of real numbers.
#   K        --  A positive integer for the rank index about which
#                VALUES should be rearranged.
# Optional:
# 
#   INDICES  --  A 1D array of original indices for elements of VALUES.
#   DIVISOR  --  A positive integer >= 2 that represents the
#                division factor used for large VALUES arrays.
#   MAX_SIZE --  An integer >= DIVISOR that represents the largest
#                sized VALUES for which the worst-case pivot value
#                selection is tolerable. A worst-case pivot causes
#                O( SIZE(VALUES)^2 ) runtime. This value should be
#                determined heuristically based on compute hardware.
# 
# Output:
# 
#   The elements of the array VALUES are rearranged such that the
#   element at position VALUES(K) is in the same location it would
#   be if all of VALUES were in sorted order. Also known as,
#   VALUES(K) has rank K.
# 
def argselect(values, k, indices=None, divisor=None, max_size=None):
    if (indices is None): indices = np.arange(len(values))
    return fast_select.argselect(values, indices, k+1, divisor=divisor, max_size=max_size)


# Class for constructing a ball tree.
class BallTree:
    # Given points and a leaf size, construct a ball tree.
    def __init__(self, points=None, transpose=True, build=True,
                 leaf_size=1, num_threads=None, max_levels=None,
                 max_copy_bytes=None, **build_kwargs):
        # Set the internals.
        self.leaf_size = leaf_size
        self.built = 0
        self.size = 0
        self.tree = None
        self.usage = None
        self.sq_sums = None
        self.order = None
        self.radii = None
        self.medians = None
        # Assign the various pruning functions.
        self._balltree = ball_tree
        self._inner = prune.inner
        self._outer = prune.outer
        self._top = prune.top
        # Configure the OpenMP environment.
        self._balltree.configure(num_threads, max_levels)
        # Set the maximum number of bytes 
        if (max_copy_bytes is not None):
            self._balltree.max_copy_bytes = max_copy_bytes
        # Assign the points and build the tree, if provided.
        if (points is not None):
            self.add(points, transpose=transpose)
            if (build): self.build(**build_kwargs)


    # Add points to this ball tree, if the type is not yet defined,
    # then initialize the type of this tree to be same as added points.
    def add(self, points, transpose=True):
        # Transpose the points if the are not already column-vectors.
        if transpose: points = points.T
        # If there are existing points, make sure the new ones match.
        if (self.tree is not None):
            # If possible, pack the points into available space.
            if (points.shape[1] <= self.tree.shape[1] - self.size):
                self.tree[:,self.size:self.size+points.shape[1]] = points
                self.size += points.shape[1]
            else:
                # Store the old internal tree-defining attributes.
                old_tree = self.tree[:,:self.size]
                old_usage = self.usage[:self.size]
                old_sq_sums = self.sq_sums[:self.size]
                old_order = self.order[:self.size]
                old_radii = self.radii[:self.built]
                old_medians = self.medians[:self.built]
                # Compute the new size and initialize the memory for the new tree.
                to_allocate = max(2 * self.size, self.size + points.shape[1])
                self.size += points.shape[1]
                self.tree = np.zeros((self.tree.shape[0], to_allocate),
                                     dtype='float32', order='F')
                # Assign the relevant internals for this tree.
                self.usage = np.zeros(self.tree.shape[1], dtype='int64')
                self.sq_sums = np.zeros(self.tree.shape[1], dtype='float32')
                self.order = np.arange(self.tree.shape[1], dtype='int64') + 1
                self.radii = np.zeros(self.tree.shape[1], dtype='float32')
                self.medians = np.zeros(self.tree.shape[1], dtype='float32')
                # Pack the old points and the new points into a single tree
                #  while ensuring the built parts of the previous tree are still valid.
                self.tree[:,:old_tree.shape[1]] = old_tree
                self.tree[:,old_tree.shape[1]:self.size] = points
                self.usage[:old_usage.size] = old_usage
                self.sq_sums[:old_sq_sums.size] = old_sq_sums
                self.order[:old_order.size] = old_order
                self.radii[:old_radii.size] = old_radii
                self.medians[:old_medians.size] = old_medians
                # Delete the old tree components.
                del old_tree, old_usage, old_sq_sums, old_order, old_radii, old_medians
        else:
            # Save the points internally as the tree.
            self.tree = np.asarray(points, order='F', dtype='float32')
            # Assign the relevant internals for this tree.
            self.usage = np.zeros(self.tree.shape[1], dtype='int64')
            self.sq_sums = np.zeros(self.tree.shape[1], dtype='float32')
            self.order = np.arange(1, self.tree.shape[1]+1, dtype='int64')
            self.radii = np.zeros(self.tree.shape[1], dtype='float32')
            self.medians = np.zeros(self.tree.shape[1], dtype='float32')
            # Store BallTree internals for knowing how to evaluate.
            self.built = 0
            self.size = self.tree.shape[1]
        # Compute the square sums for the new points.
        self._balltree.compute_square_sums(
            self.tree[:,self.built:self.size],
            self.sq_sums[self.built:self.size]
        )

    # Restructure the ball tree so the points are in locally
    # contiguous blocks of memory (local by branch + leaf).
    def reorder(self):
        self._balltree.fix_order(self.tree, self.sq_sums, self.radii, self.medians, self.order[:self.built])

    # Build a tree out.
    def build(self, leaf_size=None, root=None, reorder=None):
        # Get the leaf size if it was not given.
        if (leaf_size is not None): self.leaf_size = leaf_size
        # Translate the root from python index to fortran index if provided.
        if (root is not None): root += 1
        # After this function completes, all points will be built.
        self.built = self.size
        # Only call the routines if the tree has positive size.
        if (self.size > 0):
            # Build tree (in-place operation).
            #    .tree     will not be modified
            #    .sq_sums  will contain the squared sums of each point
            #    .radii    will be modified to have the radius of specific
            #              node, or 0.0 if this point is a leaf.
            #    .medians  will be modified to have the distance to the
            #              median child node, or 0.0 if this point is a leaf.
            #    .order    will be the list of indices (1-indexed) that
            #              determine the structure of the ball tree.
            sq_dists = np.zeros(self.size, dtype='float32')
            self._balltree.build_tree(
                self.tree, self.sq_sums,self.radii, self.medians, sq_dists,
                self.order[:self.size], leaf_size=self.leaf_size, root=root,
            )
            # Restructure the ball tree so the points are in locally contiguous blocks of
            # memory (local by branch + leaf), as long as allowed (by user or memory).
            if (reorder or ((reorder is None) and (self.tree.nbytes < self._balltree.max_copy_bytes))):
                self._balltree.fix_order(self.tree, self.sq_sums, self.radii, self.medians, self.order[:self.built])

    # Find the "k" nearest neighbors to all points in z.
    def nearest(self, z, k=1, return_distance=True, transpose=True,
                budget=None, randomness=None, include_unbuilt=True):
        # If only a single point was given, convert it to a matrix.
        if (len(z.shape) == 1):
            singleton = True
            z = z.reshape((1,-1))
        else:
            singleton = False
        # Transpose the points if appropriate.
        if transpose: z = z.T
        # Make sure the 'k' value isn't bigger than the tree size.
        k = min(k, self.size)
        n = z.shape[1]
        # Initialize holders for output.
        points = np.asarray(z, order='F', dtype='float32')
        indices = np.ones((k, points.shape[1]), order='F', dtype='int64')
        dists = np.ones((k, points.shape[1]), order='F', dtype='float32')
        iwork = np.ones((k+self.leaf_size+2, min(n,self._balltree.number_of_threads)), order='F', dtype='int64')
        rwork = np.ones((k+self.leaf_size+2, min(n,self._balltree.number_of_threads)), order='F', dtype='float32')
        # Compute the nearest neighbors.
        self._balltree.nearest(
            points, k,
            self.tree, self.sq_sums, self.radii, self.medians, self.order[:self.built],
            self.leaf_size, indices, dists, iwork, rwork, to_search=budget, randomness=randomness
        )
        del iwork, rwork
        if (budget is not None):
            budget = max(0, budget - self.built)
        # Compute distance to points that have not been built into the tree.
        if (self.size > self.built):
            uindices = np.zeros((k, points.shape[1]), order='F', dtype='int64')
            udists = np.zeros((k, points.shape[1]), order='F', dtype='float32')
            uorder = np.arange(self.size-self.built, dtype='int64') + 1
            uleaf = self.size - self.built # The "leaf size" for the unbuilt points.
            iwork = np.ones((k+uleaf+2, min(n,self._balltree.number_of_threads)), order='F', dtype='int64')
            rwork = np.ones((k+uleaf+2, min(n,self._balltree.number_of_threads)), order='F', dtype='float32')
            self._balltree.nearest(
                points, k,
                self.tree[:,self.built:], self.sq_sums[self.built:], self.radii, self.medians,
                uorder, uleaf, uindices, udists, iwork, rwork, to_search=budget,
            )
            del iwork, rwork
            # Keep the nearest of the provided dists and indices.
            indices = np.concatenate((indices, uindices+self.built), axis=0)
            dists = np.concatenate((dists, udists), axis=0)
            to_keep = np.argsort(dists, axis=0)
            for i in range(points.shape[1]):
                indices[:k,i] = indices[to_keep[:k,i],i]
                dists[:k,i] = dists[to_keep[:k,i],i]
            # Only keep the top k.
            indices = indices[:k,:]
            dists = dists[:k,:]
        # Update the usage statistics for all points referenced in return values.
        self._balltree.bincount(indices.flatten(), self.usage)
        # Revert to singleton if that's what was provided.
        if singleton:
            dists = dists[:,0]
            indices = indices[:,0]
        # Return the results.
        if return_distance:
            if transpose: return dists.T, indices.T - 1
            else:         return dists,   indices - 1
        else:
            if transpose: return indices.T - 1
            else:         return indices - 1

    # Find the "k" nearest neighbors to all points in z. Uses the same
    # interface as the "BallTree.nearest" function, see help for more info.
    def query(self, *args, **kwargs): return self.nearest(*args, **kwargs)
    def __call__(self, *args, **kwargs): return self.nearest(*args, **kwargs)

    # Summarize this tree.
    def __str__(self):
        if not hasattr(self,"tree"): return "empty BallTree"
        return f"BallTree {self.tree.shape[::-1]}[:{self.size}] -- {self.built} built"

    # Return the usable length of this tree.
    def __len__(self):
        return self.size

    # Get an index point from the tree.
    def __getitem__(self, index):
        if (self.tree is not None):
            return self.tree[:,self.order[:self.built][index]-1]
        else:
            raise(IndexError("Attempting to __getitem__ with '[]' from an empty tree."))

    # Prune this tree and compact its points into the front of the
    # array, adjust '.size' and '.built' accordingly.
    def prune(self, levels=1, full=True, build=True, method="root"):
        assert(levels >= 1)
        assert(len(self) > 0)
        # Build the tree out if it is not fully built.
        if (full and (not (self.built == self.size))): self.build()
        # Get the size of the built portion of the tree.
        size = self.built
        # Compute the indices of the inner children (50th percentile) points.
        if (levels == 1):
            # Use the middle child for 1 level.
            indices = np.array([1], dtype=np.int64)
        else:
            # Handle the different methods of pruning the tree.
            if method in {"inner", "outer"}:
                to_keep = min(size, 1 + 2**(levels-1))
                # Otherwise, get the root, first outer, and all inners.
                indices = np.zeros(to_keep, dtype=np.int64)
                # Get the indices of the inner children of the built tree.
                if method == "inner":
                    indices[0] = 1
                    indices[1] = min((size + 2) // 2, size-1) + 1
                    self._inner(size, levels, indices[2:])
                elif method == "outer":
                    indices[0] = 1
                    indices[1] = 2
                    self._outer(size, levels, indices[2:])
            else:
                to_keep = min(size, 2**levels - 1)
                # Simply grab the root of the tree.
                indices = np.zeros(to_keep, dtype=np.int64)
                self._top(size, levels, indices)
            indices[:] -= 1
        # Stop this operation if the tree will remain unchanged.
        if (len(indices) == size): return
        # Adjust the recorded size and amount built of this tree.
        self.size = len(indices)
        self.built = 1
        # Keep those indices only.
        self.order[:self.size] = self.order[indices]
        # Rebuild the tree if desired.
        if build: self.build(root=0)


if __name__ == "__main__":
    print("_"*70)
    print(" TESTING BALL TREE MODULE")

    from tlux.plot import Plot
    from tlux.random import well_spaced_ball, well_spaced_box

    # A function for testing approximation algorithms.
    def f(x):
        x = x.reshape((-1,2))
        x, y = x[:,0], x[:,1]
        return (3*x + np.cos(8*x)/2 + np.sin(5*y))

    # Set the number of points, dimension, and random seed for repeatability.
    n = 200
    d = 2
    seed = 0
    np.random.seed(seed)

    # Create the test plot.
    x = np.asarray(well_spaced_box(n, d), order="C")
    y = f(x)

    # Create the tree.
    tree = BallTree(x, reorder=False)

    # Define a function that reproduces the function via application of nearest neighbor.
    fy = lambda x: y[tree(x, return_distance=False, k=1, budget=200)]
    x_min_max = np.asarray([np.min(x,axis=0), np.max(x,axis=0)]).T

    # Generate a plot to visualize the result.
    p = Plot()
    p.add("data", *(x.T), y)
    p.add_function("balltree", fy, *x_min_max, plot_points=5000, vectorized=True)
    p.show()




# Local descriptive exceptions
class NameTypeMismatch(Exception): pass
class BadSpecifiedType(Exception): pass
class BadSpecifiedName(Exception): pass
class NoNamesSpecified(Exception): pass
class ImproperUsage(Exception): pass
class MissingModule(Exception): pass
class UnknownName(Exception): pass
class Unsupported(Exception): pass
class BadAssignment(Exception): pass
class BadElement(Exception): pass
class BadTarget(Exception): pass
class BadValue(Exception): pass
class BadIndex(Exception): pass
class BadData(Exception): pass
class Empty(Exception): pass



import numpy as np
from tlux.approximate.axy import AXY

print("_"*70)
print(" TESTING AXY MODULE")

# ----------------------------------------------------------------
#  Enable debugging option "-fcheck=bounds".
# import fmodpy
# fmodpy.config.f_compiler_args = "-fPIC -shared -O3 -fcheck=bounds"
# fmodpy.config.link_blas = ""
# fmodpy.config.link_lapack = ""
# fmodpy.config.link_blas = "-framework Accelerate"
# fmodpy.config.link_lapack = "-framework Accelerate"
# fmodpy.config.link_omp = ""
# ----------------------------------------------------------------

from tlux.plot import Plot
from tlux.random import well_spaced_ball, well_spaced_box

# TODO: test saving and loading with unique value maps
# TODO: design concise test function that has meaningful signal
#       in each of "ax", "axi", "x", "xi", test all combinations

# A function for testing approximation algorithms.
def f(x):
    x = x.reshape((-1,2))
    x, y = x[:,0], x[:,1]
    return (3*x + np.cos(8*x)/2 + np.sin(5*y))

n = 100
seed = 2
state_dim = 32
num_states = 8
steps = 1000
num_threads = None
np.random.seed(seed)
axy_kwargs = dict(
    num_threads = num_threads,
    seed = seed,
    logging_step_frequency = 1,
    rank_check_frequency = 1,
    min_update_ratio = 0.8,
    early_stop = False,
    # initial_shift_range=0.0,
    # faster_rate=1.0,
    # slower_rate=1.0,
    # basis_replacement = True,
    # orthogonalizing_step_frequency = 200,
    # keep_best = False,
    # step_replacement = 0.00,
    # equalize_y = True,
    # discontinuity=-1000.0
    # initial_step=0.01
)


TEST_FIT_SIZE = False
TEST_WEIGHTING = False
TEST_SAVE_LOAD = False
TEST_INT_INPUT = False
TEST_AGGREGATE = False
TEST_LARGE_MODEL = True
TEST_BAD_NUMBERS = False # Nan and Inf
TEST_SCALING = False # Very small (10^{-30}) and very large (10^{30}) values at once.
SHOW_VISUALS = True

if TEST_FIT_SIZE:
    # Aggreagtor model settings.
    dim_a_numeric = 4
    dim_a_embedding = None
    num_a_embeddings = 32
    dim_a_state = 8
    num_a_states = 2
    dim_a_out = None
    # Model settings.
    dim_m_numeric = 8
    dim_m_embedding = 32
    num_m_embeddings = 32
    dim_m_state = 64
    num_m_states = 8
    dim_m_out = 1
    # Number of points going into each model.
    na = 100000000
    nm = 100000
    # Initialize the model and its fit configuration.
    m = AXY(
        adn=dim_a_numeric, ade=dim_a_embedding, ane=num_a_embeddings,
        ads=dim_a_state, ans=num_a_states, ado=dim_a_out,
        mdn=dim_m_numeric, mde=dim_m_embedding, mne=num_m_embeddings,
        mds=dim_m_state, mns=num_m_states, mdo=dim_m_out,
        **axy_kwargs
    )
    print(m.config)    
    m.AXY.new_fit_config(nm, na, m.config)
    print()
    print(m)
    print(m.config)


if (not SHOW_VISUALS):
    class Plot:
        def __init__(self, *args, **kwargs): pass
        def __getattr__(self, *args, **kwargs):
            return lambda *args, **kwargs: None


if TEST_SAVE_LOAD:
    # Try saving an untrained model.
    m = AXY()
    print("Empty model:")
    print("  str(model) =", str(m))
    print()
    m.save("testing_empty_save.json")
    m.load("testing_empty_save.json")
    from tlux.approximate import AXY
    m = AXY(mdn=2, mds=state_dim, mns=num_states, mdo=1, 
            steps=steps, **axy_kwargs,
    )
    print("Initialized model:")
    print(m)
    print()
    # Create the test plot.
    x = np.asarray(well_spaced_box(n, 2), dtype="float32", order="C")
    y = f(x).astype("float32")
    # Rescale the data to make it cover many orders of magnitude.
    if TEST_SCALING:
        x[:,0] *= 10.**30
        y[:] /= 10.**30
    # Add some invalid values to the data.
    if TEST_BAD_NUMBERS:
        x[len(x)//7 : 2*len(x)//7,0] = float('nan')
        x[-2:,1] = float('inf')
        y[len(y)//7 : 2*len(y)//7] = float('nan')
        y[-2:] = float('inf')
    # Construct weights.
    yw = np.random.random(size=n) ** 5
    yw = np.where(yw > 0.5, yw*2, 0.001)
    if (not TEST_WEIGHTING): yw[:] = 1.0
    # Fit the model.
    x_normalized = x.copy()
    y_normalized = y.copy()
    m.fit(x=x_normalized, y=y_normalized, yw=yw)
    # Do mean-substitution on the invalid numbers in the dataset.
    if TEST_BAD_NUMBERS:
        x = np.where(np.isfinite(x), x, np.nan)
        x[np.isnan(x)] = (np.nanmean(x, axis=0) * np.ones(x.shape))[np.isnan(x)]
        y = np.where(np.isfinite(y), y, np.nan)
        y[np.isnan(y)] = (np.nanmean(y, axis=0) * np.ones(y.shape))[np.isnan(y)]
    # Add the data and the surface of the model to the plot.
    p = Plot()
    x_min_max = np.asarray([x.min(axis=0), x.max(axis=0)]).T
    p.add("Data", *x[yw > 0.5].T, y[yw > 0.5])
    if (TEST_WEIGHTING):
        p.add("Data (low weight)", *x[yw <= 0.5].T, y[yw <= 0.5], color=3)
    # p.add("Normalized data", *x_fit.T, y_fit)
    p.add_func("Fit", m, *x_min_max, vectorized=True)
    # Show the normalized data.
    if TEST_BAD_NUMBERS:
        q = Plot()
        q.add("Normalized data", *x_normalized.T, y_normalized, color=4)
        q.show(append=True)
    # Try saving the trained model and applying it after loading.
    print("Saving model:")
    print(m)
    print()
    m.save("testing_real_save.json")
    m.load("testing_real_save.json")
    print("Loaded model:")
    print(m)
    print()
    my = m(x.copy())[:,0]
    my_lift = (my.max() - my.min()) * 0.025
    p.add("Loaded values", *x.T, my+my_lift, color=1, marker_size=4)
    p.plot(show=(m.record.size == 0))
    # Remove the save files.
    import os
    try: os.remove("testing_empty_save.json")
    except: pass
    try: os.remove("testing_real_save.json")
    except: pass


if TEST_INT_INPUT:
    print("Building model..")
    x = well_spaced_box(n, 2)
    x_min_max = np.vstack((np.min(x,axis=0), np.max(x, axis=0))).T
    y = f(x)
    # Initialize a new model.
    m = AXY(mdn=2, mds=state_dim, mns=num_states, mdo=1, mde=3, mne=2, steps=steps, **axy_kwargs)
    all_x = np.concatenate((x, x), axis=0)
    all_y = np.concatenate((y, np.cos(np.linalg.norm(x,axis=1))), axis=0)
    all_xi = np.concatenate((np.ones(len(x)),2*np.ones(len(x)))).reshape((-1,1)).astype("int32")
    x_fit = np.array(all_x, dtype="float32", order="C")
    m.fit(x=x_fit, xi=all_xi, y=all_y.copy())
    # Create an evaluation set that evaluates the model that was built over two differnt functions.
    xi1 = np.ones((len(x),1),dtype="int32")
    y1 = m(x, xi=xi1)
    y2 = m(x, xi=2*xi1)
    print("Adding to plot..")
    p = Plot()
    # p.add("x fit", *x_fit.T, all_y[:], color=0)
    p.add("xi=1 true", *x.T, all_y[:len(all_y)//2], color=0)
    p.add("xi=2 true", *x.T, all_y[len(all_y)//2:], color=1)
    p.add_func("xi=1", lambda x: m(x.copy(), xi=np.ones(len(x), dtype="int32").reshape((-1,1))), *x_min_max, vectorized=True, color=3, shade=True)
    p.add_func("xi=2", lambda x: m(x.copy(), xi=2*np.ones(len(x), dtype="int32").reshape((-1,1))), *x_min_max, vectorized=True, color=2, shade=True)
    # Generate the visual.
    print("Generating surface plot..")
    p.show(show=False)


if TEST_AGGREGATE:
    print("Building model..")
    x = well_spaced_box(n, 2)
    x_min_max = np.vstack((np.min(x,axis=0), np.max(x, axis=0))).T
    y = f(x)
    # Create all data.
    all_x = np.concatenate((x, x), axis=0)
    all_xi = np.concatenate((np.ones(len(x)),2*np.ones(len(x)))).reshape((-1,1)).astype("int32")
    ax = all_x.reshape((-1,1)).copy()
    axi = (np.ones(all_x.shape, dtype="int32") * (np.arange(all_x.shape[1])+1)).reshape(-1,1)
    sizes = np.ones(all_x.shape[0], dtype="int32") * 2
    all_y = np.concatenate((y, np.cos(np.linalg.norm(x,axis=1))), axis=0)
    all_y = all_y.reshape((all_y.shape[0],-1))
    # Initialize a new model.
    print("Fitting model..")
    m = AXY(
        mdn=0, adn=ax.shape[1], ado=2, mdo=all_y.shape[1], 
        ads=state_dim, ans=num_states, mds=state_dim, mns=num_states,
        ane=len(np.unique(axi.flatten())), mne=len(np.unique(all_xi.flatten())),
        **axy_kwargs,
    )
    m.fit(ax=ax.copy(), axi=axi, sizes=sizes, xi=all_xi, y=all_y.copy(), steps=steps)
    # Create an evaluation set that evaluates the model that was built over two differnt functions.
    xi1 = np.ones((len(x),1),dtype="int32")
    ax = x.reshape((-1,1)).copy()
    axi = (np.ones(x.shape, dtype="int32") * (np.arange(x.shape[1])+1)).reshape(-1,1)
    sizes = np.ones(x.shape[0], dtype="int32") * 2
    temp_x = np.zeros((x.shape[0],0), dtype="float32")
    y1 = m(x=temp_x, xi=xi1, ax=ax, axi=axi, sizes=sizes)
    y2 = m(x=temp_x, xi=2*xi1, ax=ax, axi=axi, sizes=sizes)
    print("Adding to plot..")
    p = Plot()
    p.add("xi=1 true", *x.T, all_y[:len(all_y)//2,0], color=0, group=0)
    p.add("xi=2 true", *x.T, all_y[len(all_y)//2:,0], color=1, group=1)
    def fhat(x, i=1):
        xi = i * np.ones((len(x),1),dtype="int32")
        ax = x.reshape((-1,1)).copy()
        axi = (np.ones(x.shape, dtype="int32") * (np.arange(x.shape[1])+1)).reshape(-1,1)
        sizes = np.ones(x.shape[0], dtype="int32") * 2
        temp_x = np.zeros((x.shape[0],0), dtype="float32")
        return m(x=temp_x, xi=xi, ax=ax, axi=axi, sizes=sizes)
    p.add_func("xi=1", lambda x: fhat(x, 1), [0,1], [0,1], vectorized=True, color=3, opacity=0.8, group=0) #, mode="markers", shade=True)
    p.add_func("xi=2", lambda x: fhat(x, 2), [0,1], [0,1], vectorized=True, color=2, opacity=0.8, group=1) #, mode="markers", shade=True)
    # Generate the visual.
    print("Generating surface plot..")
    p.show(show=False)


if (TEST_LARGE_MODEL):  # Takes ~40 seconds on 14" 2021 M1 Pro Macbook Pro, 32GB of RAM
    # Data size.
    nm = 100000
    na = 6000000
    adn = 10
    ane = 10
    mdn = 10
    mne = 10
    mdo = 10
    # Data to fit.
    ax = np.random.random(size=(na,adn)).astype("float32")
    axi = np.random.randint(0,ane,size=(na,1)).astype("float32")
    x = np.random.random(size=(nm,mdn)).astype("float32")
    xi = np.random.randint(0,mne,size=(nm,1)).astype("float32")
    y = np.random.random(size=(nm,mdo)).astype("float32")
    # Generate random starting indices for all the sizes and
    #   convert those starting indices into sizes of batches.
    sizes = np.arange(na, dtype="int32")
    np.random.shuffle(sizes)
    sizes = sizes[:nm]
    sizes[0] = 0
    sizes.sort()
    sizes[:-1] = sizes[1:] - sizes[:-1]
    sizes[-1] = na - sizes[-1]
    # Model settings.
    ans = 4
    ads = 64
    mns = 8
    mds = 64
    steps = 11
    kwargs = axy_kwargs.copy()
    kwargs.update(dict(
        rank_check_frequency = 0,
    ))
    # Fit model.
    m = AXY(adn=adn, ane=ane, mdn=mdn, mne=mne, mdo=mdo,
             ans=ans, ads=ads, mns=mns, mds=mds, **kwargs)
    print()
    print("Fitting large model..")
    m.fit(ax=ax, axi=axi, sizes=sizes, x=x, xi=xi, y=y,
          steps=steps, **kwargs)
    print()

# Generate a visual of the loss function.
if (SHOW_VISUALS and (len(getattr(globals().get("m",None), "record", [])) > 0)):
    print()
    print("Generating loss plot..")
    p = Plot("Mean squared error")
    # Rescale the columns of the record for visualization.
    record = m.record
    for i in range(0, record.shape[0], max(1,record.shape[0] // 100)):
        step_indices = list(range(i))
        p.add("MSE", step_indices, record[:i,0], color=1, mode="lines", frame=i)
        p.add("Step factors", step_indices, record[:i,1], color=2, mode="lines", frame=i)
        p.add("Step sizes", step_indices, record[:i,2], color=3, mode="lines", frame=i)
        p.add("Update ratio", step_indices, record[:i,3], color=4, mode="lines", frame=i)
        p.add("Eval utilization", step_indices, record[:i,4], color=5, mode="lines", frame=i)
        p.add("Grad utilization", step_indices, record[:i,5], color=6, mode="lines", frame=i)
    p.show(append=True, show=True, y_range=[-.2, 1.2])
print("", "done.", flush=True)

if ("m" in globals()):
    print()
    print(m)

